package bitsplease.mummymayhem;

import bitsplease.mummymayhem.model.GameModel;
import bitsplease.mummymayhem.model.Player;
import bitsplease.mummymayhem.model.buildings.Tower;
import bitsplease.mummymayhem.model.units.Unit;
import bitsplease.mummymayhem.model.utils.SFXStorage;
import bitsplease.mummymayhem.model.utils.SFXType;
import javafx.animation.AnimationTimer;

import java.util.Random;

public class GameLoop extends AnimationTimer {
    private static final long TARGET_FRAMERATE = 144;
    private static final long MOVEMENT_UPDATE_TIME = 1_000_000_000 / TARGET_FRAMERATE;
    private static final long ANIMATION_UPDATE_TIME = 333_333_333;
    private static final long PRIORITY_THRESHOLD_UPDATE_TIME = 1_500_000_000;
    private static final double MOVEMENT_DISTANCE_PER_SECOND = 0.3;
    private static final long PROJECTILE_UPDATE_TIME = 1_000_000_000 / TARGET_FRAMERATE;
    private static final double PROJECTILE_SPEED = 10.0;

    GameModel model;
    GameViewController gameViewController;

    private int priorityThreshold = 0;
    private long lastPriorityThresholdUpdate = 0;

    public GameLoop(GameModel model, GameViewController gameViewController) {
        this.model = model;
        this.gameViewController = gameViewController;
    }

    /**
     * Updates the status of the game in every frame.
     *
     * @param l The timestamp of the current frame given in nanoseconds
     */
    @Override
    public void handle(long l) {

        updateUnits(model.getPlayerA(), l);
        updateUnits(model.getPlayerB(), l);
        updateTower(model.getPlayerA(), l);
        updateTower(model.getPlayerB(), l);

        gameViewController.drawGraphics();
        if(model.getPlayerA().getUnits().size() == 0 && model.getPlayerB().getUnits().size() == 0) {
            gameViewController.nextButtonClicked();
            priorityThreshold = 0;
        }
    }

    /**
     * Moves the selected unit according to which direction it faces.
     *
     * @param u The unit to be moved
     * @param l Time passed since last movement in nanoseconds
     */
    private void moveUnit(Unit u, long l) {
        double movementCorrection = (double)l/(double)MOVEMENT_UPDATE_TIME;
        double movementUpdateTimeSeconds = (double)MOVEMENT_UPDATE_TIME/1_000_000_000;
        double distance = MOVEMENT_DISTANCE_PER_SECOND * movementUpdateTimeSeconds * u.getSpeed();

        switch(u.getFacing()) {
            case UP -> u.moveY(movementCorrection * -(distance));
            case DOWN -> u.moveY(movementCorrection * (distance));
            case RIGHT -> u.moveX(movementCorrection * (distance));
            case LEFT -> u.moveX(movementCorrection * -(distance));
        }

        playUnitVoiceSFX();
    }

    /**
     * Updates the selected player's units' animation and location in set intervals
     *
     * @param player The selected player
     * @param l Current timestamp in nanoseconds
     */
    private void updateUnits(Player player, long l) {
        for(Unit u: player.getUnits()) {
            if(l-lastPriorityThresholdUpdate >= PRIORITY_THRESHOLD_UPDATE_TIME) {
                lastPriorityThresholdUpdate = l;
                priorityThreshold++;
            }

            if(u.getPriority() < priorityThreshold) {
                if(l-u.getLastMovementUpdate() >= MOVEMENT_UPDATE_TIME) {
                    if(u.getFacing() == null) {
                        u.setLastMovementUpdate(l);
                    }

                    if(u.getPath().size() >= 1) {
                        u.checkNextField();
                    }
                    else if(u.isDestinationReached()){
                        if(u.getDestination().hasTower()) {
                            boolean previousDemolitionState = u.getDestination().getTower().isDemolished();
                            u.attack(u.getDestination().getTower());
                            if(!previousDemolitionState && u.getDestination().getTower().isDemolished()) {
                                SFXStorage.getSFX(SFXType.TOWER_DESTROY).play();
                            }
                        }
                        else {
                            u.attack(u.getDestination().getBuilding());
                        }

                        SFXStorage.getSFX(SFXType.UNIT_ATTACK).play();

                        player.getUnits().remove(u);
                        break;
                    }
                    moveUnit(u, l-u.getLastMovementUpdate());
                    u.setLastMovementUpdate(l);
                }

                if(l-u.getLastAnimationUpdate() >= ANIMATION_UPDATE_TIME/u.getSpeed()) {
                    u.nextAnimationPhase();
                    u.setLastAnimationUpdate(l);
                }
            }
        }
    }

    /**
     * Moves the selected tower's projectile towards a unit
     *
     * @param t The selected player
     * @param u The selected unit
     * @param l Time passed since last projectile update in nanoseconds
     */
    private void moveProjectile(Tower t, Unit u, long l) {
        if(!t.isDemolished()) {
            double deltaX = t.getProjectilePosition().getX() - u.getPosition().getX();
            double deltaY = t.getProjectilePosition().getY() - u.getPosition().getY();
            double angle = Math.atan2( deltaY, deltaX );

            double movementCorrection = (double)l/(double)PROJECTILE_UPDATE_TIME;
            double projectileUpdateTimeSeconds = (double)PROJECTILE_UPDATE_TIME/1_000_000_000;
            double distance = PROJECTILE_SPEED * projectileUpdateTimeSeconds;

            double xMove = movementCorrection * distance * Math.cos(angle);
            double yMove = movementCorrection * distance * Math.sin(angle);

            t.setProjectilePosition(t.getProjectilePosition().subtract(xMove, yMove));

            if(t.getPosition().distance(u.getPosition()) <= t.getPosition().distance(t.getProjectilePosition()) ) {
                t.attack(u);
            }
        }
    }

    /**
     * Updates the selected players' towers attacks, and projectile locations
     *
     * @param player The selected player
     * @param l Current timestamp in nanoseconds
     */
    private void updateTower(Player player, long l) {
        for(Tower t: player.getTowers()) {
            if(t.getAttackCooldown() == 0 && !t.isDemolished()) {
                Player targetPlayer;
                if(player == model.getPlayerA()) {
                    targetPlayer = model.getPlayerB();
                }
                else {
                    targetPlayer = model.getPlayerA();
                }
                t.targetUnit(targetPlayer.getUnits());
                if(t.getTarget() != null) {
                    if(t.getProjectilePosition().equals(t.getPosition())) {
                        SFXStorage.getSFX(SFXType.TOWER_ATTACK).play();
                    }

                    if(l - t.getLastProjectileUpdate() >= PROJECTILE_UPDATE_TIME) {
                        moveProjectile(t, t.getTarget(),l-t.getLastProjectileUpdate());
                    }
                    t.setLastCooldownUpdate(l);

                    if(t.getTarget() != null && t.getTarget().getHp() == 0 ) {
                        targetPlayer.getUnits().remove(t.getTarget());
                        if(t.getTarget().getDistanceFromDestination() > 0) {
                            player.addGold(t.getTarget());
                            gameViewController.updateGoldLabels();

                            if(rareDeathSFX()) {
                                SFXStorage.getSFX(SFXType.UNIT_DEATH_RARE).play();
                            } else {
                                SFXStorage.getSFX(SFXType.UNIT_DEATH).play();
                            }
                        }
                    }
                }
            }
            else {
                t.lowerAttackCooldown(l-t.getLastCooldownUpdate());
                t.setLastCooldownUpdate(l);
            }
            t.setLastProjectileUpdate(l);
        }
    }

    /**
     * Determines if the rare death sfx should be player or not.
     * @return true if a 0-99 throw lands on 0 or 1, otherwise false.
     */
    private boolean rareDeathSFX() {
        Random random = new Random();
        return random.nextInt(100) < 2;
    }

    /**
     * Plays unit voice-line on a random chance.
     */
    private void playUnitVoiceSFX() {
        Random random = new Random();
        if (random.nextInt(10000) == 0) {
            SFXStorage.getSFX(SFXType.UNIT_VOICE).play();
        }
    }
}
