package bitsplease.mummymayhem;

import bitsplease.mummymayhem.model.GameModel;
import bitsplease.mummymayhem.model.PhaseType;
import bitsplease.mummymayhem.model.Player;
import bitsplease.mummymayhem.model.PlayerType;
import bitsplease.mummymayhem.model.buildings.Barrack;
import bitsplease.mummymayhem.model.buildings.Building;
import bitsplease.mummymayhem.model.buildings.Tower;
import bitsplease.mummymayhem.model.map.Map;
import bitsplease.mummymayhem.model.persistence.CustomMapPersistence;
import bitsplease.mummymayhem.model.persistence.CustomMapSave;
import bitsplease.mummymayhem.model.units.Unit;
import bitsplease.mummymayhem.model.utils.*;
import javafx.application.Platform;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.geometry.Point2D;
import javafx.geometry.Pos;
import javafx.geometry.Rectangle2D;
import javafx.scene.Node;
import javafx.scene.Scene;
import javafx.scene.canvas.Canvas;
import javafx.scene.canvas.GraphicsContext;
import javafx.scene.control.Button;
import javafx.scene.control.ButtonType;
import javafx.scene.control.Label;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.input.MouseButton;
import javafx.scene.input.MouseEvent;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.HBox;
import javafx.scene.layout.VBox;
import javafx.scene.media.MediaPlayer;
import javafx.stage.Modality;
import javafx.stage.Popup;
import javafx.stage.Screen;
import javafx.stage.Stage;
import javafx.util.Duration;

import java.io.File;
import java.io.IOException;
import java.util.Objects;
import java.util.Optional;

public class GameViewController {
    @FXML
    private ImageView quitButton;

    @FXML
    private Button loadButton;

    @FXML
    private Popup towerPopUp;

    @FXML
    private Popup buyUnitPopUp;

    @FXML
    private Popup unitStatsPopUp;

    @FXML
    private Popup gameEndPopUp;

    @FXML
    private Canvas gameCanvas;

    @FXML
    private Label phaseLabel;

    @FXML
    private Label player1Gold;

    @FXML
    private Label player2Gold;

    @FXML
    private Button buildButton;

    @FXML
    private Button nextButton;

    private GameLoop gameLoop;
    private MediaPlayer music;
    private CustomMapPersistence persistence;
    private static GameModel model;
    private Renderer renderer;
    private GraphicsContext graphics;
    private double widthRatio, heightRatio;
    private Tower selectedTower;
    private Building selectedBuilding;
    private Unit selectedUnit;
    private Point2D selectedUnitPosition;

    Label towerTypeLabel;
    Label towerDamageLabel;
    Label towerRangeLabel;
    Label towerLevelLabel;
    Label towerHpLabel;
    Button upgradeButton;
    Button demolishButton;

    /**
     * Initializes the graphicsContext, gameModel, unit movement handler, gold- and phase labels,
     * and the height and width of the fields based on the size of the map and canvas.
      */
    @FXML
    public void initialize() {
        //Music setup
        music = new MediaPlayer(ResourceLoader.loadMusic("src/main/resources/sfx/music/gerudo_valley.mp3"));
        music.setOnEndOfMedia(() -> music.seek(Duration.ZERO));
        music.setVolume(0.4);
        music.play();

        persistence = new CustomMapPersistence();
        graphics = gameCanvas.getGraphicsContext2D();
        model = new GameModel();
        gameLoop = new GameLoop(model, this);

        updatePhaseLabel();
        updateGoldLabels();

        double mapWidth = model.getMap().getWidth();
        double mapHeight = model.getMap().getHeight();

        widthRatio = gameCanvas.getWidth() / mapWidth;
        heightRatio = gameCanvas.getHeight() / mapHeight;

        renderer = new Renderer(model.getMap(), graphics);
        renderer.setHeightRatio(heightRatio);
        renderer.setWidthRatio(widthRatio);

        drawGraphics();

        Image exitImg = ResourceLoader.loadImage("src/main/resources/assets/images/exit.png");
        quitButton.setImage(exitImg);
        registerQuitButton();
        registerLoadButton();
        updatePhaseLabel();
        updateGoldLabels();
    }

    /**
     * Handles the click event on canvas, create building/unit buying popups and selecting fields to place buildings.
     * Enters path modification mode if there is a selected unit.
     *
     * @param event Mouse Event details
     */
    @FXML
    protected void canvasClicked(MouseEvent event) {
        double row = event.getY() / gameCanvas.getHeight() * model.getMap().getHeight();
        double column = event.getX() / gameCanvas.getWidth() * model.getMap().getWidth();
        if(selectedUnit == null) {
            handleNormalMode(event, column, row);
        }
        else {
            handlePathModificationMode(event, (int)column, (int)row);
        }
    }

    /**
     * Handles click events on canvas in normal mode.
     *
     * @param event The mouse event
     * @param column The column number
     * @param row The row number
     */
    private void handleNormalMode(MouseEvent event, double column, double row) {
        hideBuyUnitPopUp();
        if (model.getSelectedBuilding() == null && model.getPhase() != PhaseType.ATTACK
                && (model.getMap().isCastlePosition((int) column, (int) row) || model.isBarrackPosition((int) column, (int) row))) {
            onCastleOrBarrackClicked(event, column, row);
        }

        if (towerPopUp == null && model.getSelectedBuilding() != null) {
            onBuildRequest(column, row);
        }
        else if (towerPopUp != null && model.getSelectedBuilding() == null
                && model.getPhase() != PhaseType.ATTACK && model.getMap().getField((int) row, (int) column).hasTower()) {
            hideTowerPopUp();
            onTowerClicked(event, column, row);
        }
        else if (towerPopUp != null) {
            hideTowerPopUp();
            if (model.getSelectedBuilding() != null) {
                onBuildRequest(column, row);
            }
        }
        else if (model.getSelectedBuilding() == null && model.getMap().getField((int) row, (int) column).hasTower()) {
            onTowerClicked(event, column, row);
        }
    }

    /**
     * Handles click events on canvas in path modification mode.
     * With primary click it modifies the selected unit's path,
     * with secondary click it exits from modification mode.
     *
     * @param event The mouse event
     * @param column The column number
     * @param row The row number
     */
    private void handlePathModificationMode(MouseEvent event, int column, int row) {
        if(event.getButton() == MouseButton.SECONDARY) {
            selectedUnit = null;
            selectedUnitPosition = null;
            drawGraphics();

            Stage thisStage = getStage(event);
            buyUnitPopUp.show(thisStage);
            updateBuildButton();
            loadButton.setDisable(false);
        }
        else if(event.getButton() == MouseButton.PRIMARY) {
            model.modifyUnitPath(selectedUnit, selectedUnitPosition, column, row);
            drawGraphics();
        }
    }

    /**
     * Handles click events on castles and barracks.
     * If it's an appropriate building and phase, creates the buyUnitPopUp.
     *
     * @param event The mouse event
     * @param column The column number
     * @param row The row number
     */
    private void onCastleOrBarrackClicked(MouseEvent event, double column, double row) {
        Stage thisStage = getStage(event);
        if ((model.getMap().getPlayerACastlePosition().equals(new Point2D((int) column, (int) row))
                || model.getPlayerA().hasBarrackOnPosition((int) column, (int) row))
                && model.getPhase() == PhaseType.PLAYER_A_BUILD) {
            createBuyUnitPopUp((int)row, (int)column, thisStage, "right_1.png", (int) column, (int) row, model.getPlayerA());
            selectedBuilding = model.getMap().getField((int)row, (int)column).getBuilding();
        }
        else if ((model.getMap().getPlayerBCastlePosition().equals(new Point2D((int) column, (int) row))
                || model.getPlayerB().hasBarrackOnPosition((int) column, (int) row))
                && model.getPhase() == PhaseType.PLAYER_B_BUILD) {
            createBuyUnitPopUp((int)row, (int)column, thisStage, "left_1.png", (int) column, (int) row, model.getPlayerB());
            selectedBuilding = model.getMap().getField((int)row, (int)column).getBuilding();
        }
    }

    /**
     * Places a tower on the desired position if the selected fields is not occupied,
     * there is no building in its surroundings,
     * the tower does not block the path between the players' castles,
     * and the player has sufficient gold to buy the barrack.
     * Updates the gold labels and renders the graphics afterwards.
     *
     * @param row The row of the desired position
     * @param column The column of the desired position
     */
    private void placeTower(double row, double column) {
        if(model.getMap().getField((int)row, (int)column).isOccupied() || !model.isValidTowerPosition((int)column, (int)row)) {
            model.selectBuilding(null);
            DialogDisplayer.displayErrorDialog("The tower is not on your own half, the field is already occupied or there's a barrack nearby!");
            return;
        }
        Tower newTower = (Tower) model.getSelectedBuilding();
        newTower.setPosition(new Point2D((int)column, (int)row));
        if(!model.getMap().blockPath(newTower)) {
            model.selectBuilding(null);
            if(model.getActualPlayer().buyTower(newTower)) {
                model.getMap().addTower((int)column, (int)row, newTower);
                model.refreshUnitsPath((int)column, (int)row);
                drawGraphics();
                updateGoldLabels();
                SFXStorage.getSFX(SFXType.TOWER_PLACEMENT).play();
            }
        } else {
            DialogDisplayer.displayErrorDialog("The tower is near the castle or would block the path!");
        }
    }

    /**
     * Places a barrack on the desired position if the selected fields is not occupied,
     * there is no building in its surroundings,
     * the barrack does not block the path between the players' castles,
     * and the player has sufficient gold to buy the barrack.
     * Updates the gold labels and renders the graphics afterwards.
     *
     * @param row The row of the desired position
     * @param column The column of the desired position
     */
    private void placeBarrack(double row, double column) {
        if(model.getMap().getField((int)row, (int)column).isOccupied()
                || !model.isValidBarrackPosition((int)column, (int)row)
                || model.playerReachedBarrackLimit()) {
            model.selectBuilding(null);
            DialogDisplayer.displayErrorDialog("The barrack is in invalid position or you've reached the barrack limit!");
            return;
        }
        Barrack newBarrack = (Barrack) model.getSelectedBuilding();
        newBarrack.setPosition(new Point2D((int)column, (int)row));
        if(!model.getMap().blockPath(newBarrack)) {
            model.selectBuilding(null);
            if(model.getActualPlayer().buyBarrack(newBarrack)) {
                model.getMap().addBuilding((int)column, (int)row, newBarrack);
                model.refreshUnitsPath((int)column, (int)row);
                drawGraphics();
                updateGoldLabels();

                SFXStorage.getSFX(SFXType.TOWER_PLACEMENT).play();
            }
        } else {
            DialogDisplayer.displayErrorDialog("The barrack is near the castle or would block the path!");
        }
    }

    /**
     * Attempts to place a tower or a barrack
     * to the given position.
     *
     * @param column The column number
     * @param row The row number
     */
    private void  onBuildRequest(double column, double row) {
        if (model.getSelectedBuilding() instanceof Tower) {
            placeTower(row, column);
        } else if (model.getSelectedBuilding() instanceof Barrack) {
            placeBarrack(row, column);
        }
    }

    /**
     * Creates the towerPopUp at the given position.
     *
     * @param event The mouse event
     * @param column The column number
     * @param row The row number
     */
    private void onTowerClicked(MouseEvent event, double column, double row) {
        selectedTower = model.getMap().getField((int) row, (int) column).getTower();
        Stage thisStage = getStage(event);
        createTowerPopUp((int)row, (int)column, thisStage);
    }

    /**
     * Creates a unit buying popup and adds event handlers to child items,
     * then shows it the given stage and sets its anchor points.
     *
     * @param x The x coordinate of the map
     * @param y The y coordinate of the map
     * @param stage The parent stage
     * @param fileName The file name of units
     * @param column The column number
     * @param row The row number
     * @param player The player who clicked
     */
    private void createBuyUnitPopUp(int x, int y, Stage stage, String fileName, int column, int row, Player player) {
        buyUnitPopUp = new Popup();
        HBox unitsListItem = new HBox();
        unitsListItem.setSpacing(10);
        unitsListItem.setPrefHeight(100);
        unitsListItem.getStyleClass().add("units-list-item");
        for (Unit unit : GameModel.units) {
            VBox unitItem = new VBox();
            unitItem.setAlignment(Pos.CENTER);
            unitsListItem.getStyleClass().add("unit-item");

            ImageView buildingImage = new ImageView();
            Image unitImage = ResourceLoader.loadImage(unit.getImageFolderPath() + fileName);
            buildingImage.setImage(unitImage);
            buildingImage.setFitWidth(50);
            buildingImage.setPreserveRatio(true);

            Label nameLabel = new Label(unit.getName());
            Label costLabel = new Label(unit.getPrice() + " g");

            int boughtCounter = model.getUnitBoughtNumber(unit, column, row);

            Label boughtLabel = new Label("Bought: " + boughtCounter);
            BorderPane borderPane = new BorderPane();

            registerUnitBorderPane(stage, borderPane, boughtLabel, column, row, unit, player);
            unitItem.getChildren().addAll(buildingImage, nameLabel, costLabel, boughtLabel);
            borderPane.setCenter(unitItem);
            unitsListItem.getChildren().add(borderPane);
        }

        buyUnitPopUp.getContent().add(unitsListItem);
        buyUnitPopUp.show(stage);
        this.setPopupAnchors(buyUnitPopUp, y, x, stage);
    }

    /**
     * Create a Tower Info popup, containing the selected tower's stats. The owner of the tower can upgrade
     * and demolish it.
     *
     * @param x the x coordinate of the map
     * @param y the y coordinate of the map
     * @param stage the parent stage
     */
    private void createTowerPopUp(int x, int y, Stage stage) {
        towerPopUp = new Popup();
        VBox towerInfoItem = new VBox();
        towerTypeLabel = new Label("Type: " + selectedTower.getType());
        towerHpLabel = new Label("Hp: " + selectedTower.getHp());
        towerDamageLabel = new Label("Damage: " + selectedTower.getAttack());
        towerRangeLabel = new Label("Range: " + selectedTower.getRange());
        towerLevelLabel = new Label("Level: " + selectedTower.getLevel());

        towerInfoItem.setPrefHeight(100);
        towerInfoItem.getStyleClass().add("towerinfo-item");
        towerInfoItem.getChildren().addAll(towerTypeLabel, towerDamageLabel, towerRangeLabel, towerLevelLabel, towerHpLabel);
        if(model.getActualPlayer().searchTower(selectedTower) && model.getPhase() != PhaseType.ATTACK) {
            HBox towerButtons = new HBox();
            upgradeButton = new Button("Upgrade");
            demolishButton = new Button("Demolish");
            towerButtons.setSpacing(15);
            upgradeButton.getStyleClass().add("action-button");
            demolishButton.getStyleClass().add("action-button");
            checkUpgradeButton();
            registerUpgradeHandlers();
            registerDemolishHandler();
            towerButtons.getChildren().addAll(upgradeButton, demolishButton);
            towerInfoItem.getChildren().add(towerButtons);
        }
        towerPopUp.getContent().add(towerInfoItem);
        towerPopUp.show(stage);
        this.setPopupAnchors(towerPopUp, y, x, stage);
    }

    /**
     * Creates a unit stats popup, shows it on the given stage
     * and sets its anchor points.
     *
     * @param stage The parent stage
     * @param children The number of the child which was clicked
     */
    private void createUnitStatsPopUp(Stage stage, int children) {
        unitStatsPopUp = new Popup();
        VBox unitStatItem = new VBox();
        Label hp = new Label("Hp: " + GameModel.units.get(children).getHp());
        Label attack = new Label("Attack: " + GameModel.units.get(children).getAttack());
        Label speed = new Label("Speed: " + GameModel.units.get(children).getSpeed());
        unitStatItem.setPrefHeight(60);
        unitStatItem.getStyleClass().add("unitstat-item");
        unitStatItem.getChildren().addAll(hp, attack, speed);

        unitStatsPopUp.getContent().add(unitStatItem);
        unitStatsPopUp.show(stage);
        unitStatsPopUp.setAnchorX(buyUnitPopUp.getAnchorX() + buyUnitPopUp.getWidth()/5 * children);
        unitStatsPopUp.setAnchorY(buyUnitPopUp.getAnchorY() - unitStatsPopUp.getHeight() - unitStatsPopUp.getHeight()/10);
    }

    /**
     * Creates the end of game popup, shows it and sets its anchor points.
     */
    private void createGameEndPopUp() {
        gameEndPopUp = new Popup();
        VBox content = new VBox();
        content.setPrefHeight(300);
        content.setPrefWidth(300);
        content.setSpacing(10);
        content.getStyleClass().add("endgame-item");
        Label winner = new Label();

        String text;
        if(model.getWinner() != null) {
            winner = new Label(model.getWinner().name() + " won!");
            if(model.getWinner().equals(PlayerType.PLAYER_A)) {
                text = "Player 1 won!";
            }
            else {
                text = "Player 2 won!";
            }
        }
        else {
            text = "DRAW";
        }
        winner.setText(text);
        winner.getStyleClass().add("winner-label");

        Button menuButton = new Button("BACK TO MAIN MENU");
        menuButton.getStyleClass().add("end-game-button");
        registerGameEndMenuButton(menuButton);

        Button exitButton = new Button("EXIT GAME");
        exitButton.getStyleClass().add("end-game-button");
        registerGameEndExitButton(exitButton);

        content.getChildren().addAll(winner, menuButton, exitButton);
        gameEndPopUp.getContent().add(content);

        Stage stage = (Stage) gameCanvas.getScene().getWindow();
        gameEndPopUp.show(stage);
        this.setPopupAnchors(gameEndPopUp, 0,0, stage);
    }

    /**
     * Handle the clicking event on Demolish button, by removing the selected tower.
     */
    private void registerDemolishHandler() {
        demolishButton.addEventHandler(MouseEvent.MOUSE_CLICKED, mouseEvent -> {
            Point2D position = selectedTower.getPosition();
            if (model.getActualPlayer().removeTower(selectedTower)) {
                model.getMap().getField((int)position.getY(), (int)position.getX()).removeTower();

                SFXStorage.getSFX(SFXType.TOWER_DEMOLISH).play();
            }
            drawGraphics();
            hideTowerPopUp();
            updateGoldLabels();
        });
    }

    /**
     * Handle different mouse events on Upgrade Button. Entering the button, we see how the tower would upgrade after
     * actually upgrading it. Exiting it, stats go back. Clicking it, the tower truly upgrade.
     */
    protected void registerUpgradeHandlers() {
        upgradeButton.addEventHandler(MouseEvent.MOUSE_CLICKED, mouseEvent -> {
            model.getActualPlayer().upgradeTower(selectedTower);
            updateGoldLabels();
            setPopUpToDefault();
            checkUpgradeButton();

            SFXStorage.getSFX(SFXType.TOWER_UPGRADE).play();
        });
        upgradeButton.addEventHandler(MouseEvent.MOUSE_ENTERED, mouseEvent -> updatePopUp());
        upgradeButton.addEventHandler(MouseEvent.MOUSE_EXITED, mouseEvent -> setPopUpToDefault());
    }

    /**
     * Handles mouse events on loadButton.
     */
    private void registerLoadButton() {
        loadButton.addEventHandler(MouseEvent.MOUSE_CLICKED, mouseEvent -> {
            SFXStorage.getSFX(SFXType.BUTTON_CLICK).play();
            hideTowerPopUp();
            hideBuyUnitPopUp();
            Stage thisStage = getStage(mouseEvent);
            File file = DialogDisplayer.displayLoadingWindow(thisStage);
            if(file != null) {
                String path = file.getPath();
                CustomMapSave result = persistence.load(path);
                if(!result.getFields().isEmpty()) {
                    Map customMap = new Map(result.getFields());
                    model = new GameModel(customMap);
                    refreshController();
                }
            }
        });
    }

    /**
     * Handles mouse events on quitButton.
     */
    private void registerQuitButton() {
        quitButton.addEventHandler(MouseEvent.MOUSE_CLICKED, mouseEvent -> {
            SFXStorage.getSFX(SFXType.BUTTON_CLICK).play();
            Optional<ButtonType> result = DialogDisplayer.displayConfirmationDialog();
            if (result.isPresent() && result.get().getText().equals("Yes")) {
                resetPopUps();
                Stage thisStage = getStage(mouseEvent);
                fxmlLoader(thisStage, "main-window.fxml");
                music.stop();
                gameLoop.stop();
            }
        });
    }

    /**
     * Handles mouse events on unit's BorderPane.
     * The click events are responsible for unit buying,
     * unit selling and path modification mode's start.
     * The mouse over events are responsible for showing the unit's stats.
     *
     * @param stage The current stage
     * @param borderPane The unit's border pane
     * @param boughtLabel The unit's bought label
     * @param column The column number
     * @param row The row number
     * @param unit The unit sample
     * @param player The current player
     */
    private void registerUnitBorderPane(Stage stage, BorderPane borderPane, Label boughtLabel, int column, int row, Unit unit, Player player) {
        borderPane.addEventHandler(MouseEvent.MOUSE_CLICKED, mouseEvent -> {
            Point2D position = new Point2D(column, row);
            if(mouseEvent.getButton() == MouseButton.PRIMARY) {
                Unit boughtUnit = model.playerBuyUnit(unit, position);
                if(boughtUnit!= null) {
                    boughtUnit.initAnimationStages();
                    SFXStorage.getSFX(SFXType.UNIT_BUY).play();
                }
            } else if(mouseEvent.getButton() == MouseButton.SECONDARY) {
                if(player.sellUnit(unit, position)) {
                    SFXStorage.getSFX(SFXType.UNIT_SELL).play();
                }
            } else if(mouseEvent.getButton() == MouseButton.MIDDLE) {
                boolean selectUnit = model.getActualPlayer().getUnits().stream()
                        .filter(u -> u.getClass().equals(unit.getClass()))
                        .anyMatch(u -> u.getPosition().equals(new Point2D(column, row)));

                if (selectUnit) {
                    selectedUnit = unit;
                    selectedUnitPosition = new Point2D(column, row);
                    buyUnitPopUp.hide();
                    drawGraphics();
                    updateBuildButton();
                    loadButton.setDisable(true);
                }
            }

            int boughtCounter = model.getUnitBoughtNumber(unit, column, row);
            boughtLabel.setText("Bought: " + boughtCounter);
            updateGoldLabels();
        });

        borderPane.addEventHandler(MouseEvent.MOUSE_ENTERED, mouseEvent -> {
            hideUnitStatsPopUp();
            Node node = (Node) mouseEvent.getSource();
            int children = node.getParent().getChildrenUnmodifiable().indexOf(node);
            createUnitStatsPopUp(stage, children);
        });

        borderPane.addEventHandler(MouseEvent.MOUSE_EXITED, mouseEvent -> hideUnitStatsPopUp());
    }

    /**
     * Handles click events on the given button.
     * It redirects to the main menu.
     *
     * @param menuButton The menuButton instance
     */
    private void registerGameEndMenuButton(Button menuButton) {
        menuButton.addEventHandler(MouseEvent.MOUSE_CLICKED, mouseEvent -> {
            gameEndPopUp.hide();
            gameEndPopUp = null;
            Stage thisStage = (Stage) gameCanvas.getScene().getWindow();
            fxmlLoader(thisStage, "main-window.fxml");
        });
    }

    /**
     * Handles click events on the given button.
     * It exits the program.
     *
     * @param exitButton The exitButton instance
     */
    private void registerGameEndExitButton(Button exitButton) {
        exitButton.addEventHandler(MouseEvent.MOUSE_CLICKED, mouseEvent -> Platform.exit());
    }

    /**
     * Hides any popup and sets it to null.
     */
    private void resetPopUps() {
        if(towerPopUp != null) {
            hideTowerPopUp();
        }
        else if(buyUnitPopUp != null) {
            hideBuyUnitPopUp();
        }
        else if(unitStatsPopUp != null) {
            hideBuyUnitPopUp();
        }
        else if(gameEndPopUp != null) {
            hideGameEndPopUp();
        }
    }

    /**
     * Refresh the fields according to model's map,
     * starts a new game loop, draws the game are and update labels.
     */
    private void refreshController() {
        double mapWidth = model.getMap().getWidth();
        double mapHeight = model.getMap().getHeight();
        widthRatio = gameCanvas.getWidth() / mapWidth;
        heightRatio = gameCanvas.getHeight() / mapHeight;
        renderer.setMap(model.getMap());
        renderer.setHeightRatio(heightRatio);
        renderer.setWidthRatio(widthRatio);
        gameLoop = new GameLoop(model, this);
        drawGraphics();
        updateGoldLabels();
        updatePhaseLabel();
    }

    /**
     * Handle the click event on Build button by create the Building Dialog window.
     */
    @FXML
    protected void buildButtonClicked() {
        SFXStorage.getSFX(SFXType.BUTTON_CLICK).play();
        hideTowerPopUp();
        hideBuyUnitPopUp();
        Stage st = new Stage();
        st.initModality(Modality.APPLICATION_MODAL);
        st.setTitle("Building Dialog");
        st.getIcons().add(new Image(new File("src/main/resources/assets/images/building-dialog-icon.png").toURI().toString()));
        fxmlLoader(st, "build-dialog.fxml");
        st.showAndWait();
    }

    /**
     * Jump to the next Phase and update the Phase Label.
     * Starts the units' movement in attack phase, and stops it in building phase.
     */
    @FXML
    protected void nextButtonClicked() {
        hideTowerPopUp();
        hideBuyUnitPopUp();
        if(selectedUnit != null) {
            selectedUnit = null;
            selectedUnitPosition = null;
            drawGraphics();
        }

        PhaseType previousPhase = model.getPhase();
        if(model.getPhase().equals(PhaseType.ATTACK)){
            gameLoop.stop();

            if(!model.gameEnded()) {
                SFXStorage.getSFX(SFXType.BATTLE_PHASE_END).play();
            }
            if(model.gameEnded()) {
                buildButton.setDisable(true);
                nextButton.setDisable(true);
                loadButton.setDisable(true);
                drawGameEnd();
                playGameEndSFX();
                return;
            }
            model.getPlayerA().addNextRoundGold();
            model.getPlayerB().addNextRoundGold();
            updateGoldLabels();
            nextButton.setDisable(false);
            loadButton.setDisable(false);
            resetProjectilePositions();
            drawGraphics();
        }
        model.nextPhase();
        updatePhaseLabel();
        if(model.getPhase().equals(PhaseType.ATTACK)) {
            if(model.getPlayerB().getUnits().size()!=0 || model.getPlayerA().getUnits().size() != 0) {
                SFXStorage.getSFX(SFXType.BATTLE_PHASE_START).play();
            }
            nextButton.setDisable(true);
            loadButton.setDisable(true);
            drawGraphics();
            model.initializeUnitPriorities();
            gameLoop.start();
        }
        else if (model.getPhase().equals(PhaseType.PLAYER_A_BUILD) || model.getPhase().equals(PhaseType.PLAYER_B_BUILD)
                && previousPhase.equals(PhaseType.PLAYER_A_BUILD) || previousPhase.equals(PhaseType.PLAYER_B_BUILD)) {
            SFXStorage.getSFX(SFXType.BUTTON_CLICK).play();
        }
    }

    /**
     * Disable Upgrade button if there is not enough gold or the selected tower is demolished by a unit.
     */
    private void checkUpgradeButton() {
        if(model.getActualPlayer().getGold() < selectedTower.getPrice()*0.5 || selectedTower.isDemolished() ||
                selectedTower.getHp() == 0 || selectedTower.getLevel() == Tower.MAX_LEVEL) { upgradeButton.setDisable(true); }
    }

    /**
     * Set the Tower info popup to default, after exiting the Upgrade button.
     */
    private void setPopUpToDefault() {
        towerHpLabel.setText("Hp: " + selectedTower.getHp());
        towerHpLabel.setStyle("-fx-text-fill:black;");
        towerDamageLabel.setText("Damage: " + selectedTower.getAttack());
        towerDamageLabel.setStyle("-fx-text-fill:black;");
        towerRangeLabel.setText("Range: " + selectedTower.getRange());
        towerRangeLabel.setStyle("-fx-text-fill:black;");
        towerLevelLabel.setText("Level: " + selectedTower.getLevel());
    }

    /**
     * Update the Tower info popup, displaying the tower's stat after upgrading.
     */
    private void updatePopUp() {
        switch (selectedTower.getUpgradeType()) {
            case HP -> {
                towerHpLabel.setText("Hp: " + selectedTower.getHp() + " + 10");
                towerHpLabel.setStyle("-fx-text-fill:green;");
            }
            case ATTACK -> {
                towerDamageLabel.setText("Damage: " + selectedTower.getAttack() + " + 2");
                towerDamageLabel.setStyle("-fx-text-fill:green;");
            }
            case RANGE -> {
                towerRangeLabel.setText("Range: " + selectedTower.getRange() + " + 1");
                towerRangeLabel.setStyle("-fx-text-fill:green;");
            }
        }
    }

    /**
     * Checks the actual phase of the game,
     * and updates the label accordingly.
     */
    protected void updatePhaseLabel() {
        StringBuilder stringBuilder = new StringBuilder();
        if(model.getPhase() == PhaseType.PLAYER_A_BUILD) {
            stringBuilder.append("Building Phase: Player 1");
        }
        else if(model.getPhase() == PhaseType.PLAYER_B_BUILD) {
            stringBuilder.append("Building Phase: Player 2");
        }
        else {
            stringBuilder.append("Attack Phase");
        }
        phaseLabel.setText(stringBuilder.toString());
        updateBuildButton();
    }

    /**
     * Checks the amount of gold each player has,
     * and updates the labels accordingly.
     */
    protected void updateGoldLabels() {
        player1Gold.setText(model.getPlayerA().getGold() + " g");
        player2Gold.setText(model.getPlayerB().getGold() + " g");
    }

    /**
     * Disable Build button at Attack Phase and enable it at building phase.
     */
    private void updateBuildButton() {
        buildButton.setDisable(model.getPhase() == PhaseType.ATTACK || selectedUnit != null);
    }

    /**
     * A static method to get the model object on Building Dialog.
     * @return model object
     */
    public static GameModel getModel() { return model; }

    /**
     * Returns the given event's source stage.
     *
     * @param event The mouse event
     * @return The stage
     */
    private Stage getStage(MouseEvent event) {
        Node node = (Node) event.getSource();
        return (Stage) node.getScene().getWindow();
    }

    /**
     * Hides the buyUnitPopUp and sets the according values to null.
     */
    private void hideBuyUnitPopUp() {
        if (buyUnitPopUp != null) {
            buyUnitPopUp.hide();
            buyUnitPopUp = null;
            selectedBuilding = null;
        }
    }

    /**
     * Hides the unitStatsPopUp and sets the according values to null.
     */
    private void hideUnitStatsPopUp() {
        if (unitStatsPopUp != null) {
            unitStatsPopUp.hide();
            unitStatsPopUp = null;
        }
    }

    /**
     * Hides the towerPopUp and sets the according values to null.
     */
    private void hideTowerPopUp() {
        if(towerPopUp != null) {
            towerPopUp.hide();
            towerPopUp = null;
            selectedTower = null;
        }
    }

    /**
     * Hides the gameEndPopUp and sets the according values to null.
     */
    private void hideGameEndPopUp() {
        if (gameEndPopUp != null) {
            gameEndPopUp.hide();
            gameEndPopUp = null;
        }
    }

    /**
     * Sets the popup's position to the given canvas position (row, column),
     * or to the middle of the canvas if it's a gameEndPopup.
     *
     * @param popup The Popup instance
     * @param row The row number
     * @param column The column number
     * @param stage The window
     */
    private void setPopupAnchors(Popup popup, int row, int column, Stage stage) {
        double canvasPosX = stage.getX() + gameCanvas.getScene().getX() + gameCanvas.getBoundsInParent().getMinX();
        double canvasPosY = stage.getY() + gameCanvas.getScene().getY() + gameCanvas.getBoundsInParent().getMinY();
        if(popup == gameEndPopUp) {
            gameEndPopUp.setAnchorX(canvasPosX + gameCanvas.getWidth()/2 - gameEndPopUp.getWidth()/2);
            gameEndPopUp.setAnchorY(canvasPosY + gameCanvas.getHeight()/2 - gameEndPopUp.getHeight()/2);
        }
        else {
            popup.setAnchorX(canvasPosX + row*widthRatio - popup.getWidth()/2 + widthRatio/2);
            popup.setAnchorY(canvasPosY + column*heightRatio - popup.getHeight() - heightRatio/2);
        }
    }

    /**
     * Hides or shows the popup according to the window location.
     *
     * @param popup The Popup instance
     * @param stage The window
     */
    private void handlePopupVisibility(Popup popup, Stage stage) {
        Rectangle2D screenBounds = Screen.getPrimary().getVisualBounds();
        if(popup == gameEndPopUp) {
            double canvasPosX = stage.getX() + gameCanvas.getScene().getX() + gameCanvas.getBoundsInParent().getMinX();
            double canvasPosY = stage.getY() + gameCanvas.getScene().getY() + gameCanvas.getBoundsInParent().getMinY();
            if(canvasPosX + gameCanvas.getWidth()/2 + gameEndPopUp.getWidth()/2 >= screenBounds.getMaxX()
                    || canvasPosY + gameCanvas.getHeight()/2 + gameEndPopUp.getHeight()/2 >= screenBounds.getMaxY()
                    || canvasPosX + gameCanvas.getWidth()/2 - gameEndPopUp.getWidth()/2 <= screenBounds.getMinX()
                    || canvasPosY + gameCanvas.getHeight()/2 - gameEndPopUp.getHeight()/2 <= screenBounds.getMinY()) {
                popup.hide();
            } else {
                popup.show(stage);
            }
        }
        else {
            if(popup.getAnchorX() + popup.getWidth() >= screenBounds.getMaxX()
                    || popup.getAnchorY() + popup.getHeight() >= screenBounds.getMaxY()
                    || popup.getAnchorX() <= screenBounds.getMinX()
                    || popup.getAnchorY() <= screenBounds.getMinY()) {
                popup.hide();
            } else {
                popup.show(stage);
            }
        }
    }

    /**
     * Moves the current popup's position according to window location.
     *
     * @param stage The window
     */
    public void movePopupsWithWindow(Stage stage) {
        if (towerPopUp != null) {
            int row = (int)selectedTower.getPosition().getX();
            int column = (int)selectedTower.getPosition().getY();
            this.setPopupAnchors(towerPopUp, row, column, stage);
            handlePopupVisibility(towerPopUp, stage);
        }

        if (buyUnitPopUp != null) {
            int row = (int)selectedBuilding.getPosition().getX();
            int column = (int)selectedBuilding.getPosition().getY();
            this.setPopupAnchors(buyUnitPopUp, row, column, stage);
            handlePopupVisibility(buyUnitPopUp, stage);
        }

        if (unitStatsPopUp != null) {
            handlePopupVisibility(unitStatsPopUp, stage);
        }

        if(gameEndPopUp != null) {
            this.setPopupAnchors(gameEndPopUp, 0,0, stage);
            handlePopupVisibility(gameEndPopUp, stage);
        }
    }

    /**
     * Resets every tower's projectile position
     */
    private void resetProjectilePositions() {
        for(Tower t: model.getPlayerA().getTowers()) {
            t.resetProjectilePosition();
        }
        for(Tower t: model.getPlayerB().getTowers()) {
            t.resetProjectilePosition();
        }
    }

    /**
     * Renders the graphics of the game
     */
    public void drawGraphics() {
        graphics.clearRect(0,0, gameCanvas.getWidth(), gameCanvas.getHeight());
        renderer.drawFields();
        renderer.drawUnitPath(model, selectedUnit, selectedUnitPosition);
        renderer.drawCastles();
        renderer.drawTowers(model);
        renderer.drawBarracks(model);
        renderer.drawHpBars(model);
        renderer.drawUnits(model);
    }

    /**
     * Darkens the canvas and creates the gameEndPopUp.
     */
    public void drawGameEnd() {
        renderer.darkenCanvas();
        createGameEndPopUp();
        music.stop();
    }

    /**
     * Plays the correct SFX depending on the outcome of the game.
     */
    private void playGameEndSFX() {
        if(!model.getPlayerA().getCastle().isAlive() && !model.getPlayerB().getCastle().isAlive()) {
            SFXStorage.getSFX(SFXType.GAME_END_DRAW).play();
        } else {
            SFXStorage.getSFX(SFXType.GAME_END_VICTORY).play();
        }
    }

    /**
     * Loads a scene from the given path.
     *
     * @param thisStage The current stage
     * @param file The scene path
     */
    private void fxmlLoader(Stage thisStage, String file) {
        FXMLLoader fxmlLoader = new FXMLLoader(getClass().getResource(file));
        Scene gameScene = null;
        try {
            gameScene = new Scene(fxmlLoader.load());
        } catch (IOException e) {
            DialogDisplayer.displayErrorDialog("The scene loading was unsuccessful!");
        }
        Objects.requireNonNull(gameScene).getStylesheets().add(Objects.requireNonNull(getClass().getResource("mmstyle.css")).toExternalForm());
        thisStage.setScene(gameScene);
    }
}