package bitsplease.mummymayhem;

import bitsplease.mummymayhem.model.map.Field;
import bitsplease.mummymayhem.model.map.FieldType;
import bitsplease.mummymayhem.model.map.Map;
import bitsplease.mummymayhem.model.persistence.CustomMapPersistence;
import bitsplease.mummymayhem.model.persistence.CustomMapSave;
import bitsplease.mummymayhem.model.utils.*;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.geometry.Pos;
import javafx.geometry.Rectangle2D;
import javafx.scene.Node;
import javafx.scene.Scene;
import javafx.scene.canvas.Canvas;
import javafx.scene.canvas.GraphicsContext;
import javafx.scene.control.Button;
import javafx.scene.control.ButtonType;
import javafx.scene.control.Label;
import javafx.scene.control.TextField;
import javafx.scene.image.ImageView;
import javafx.scene.input.MouseEvent;
import javafx.scene.layout.HBox;
import javafx.scene.layout.VBox;
import javafx.scene.media.MediaPlayer;
import javafx.stage.Popup;
import javafx.stage.Screen;
import javafx.stage.Stage;
import javafx.util.Duration;

import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Objects;
import java.util.Optional;

public class MapEditorController {
    @FXML
    private Popup fieldTypeSelectingPopup;

    @FXML
    private VBox settings;

    @FXML
    Canvas mapEditorCanvas;

    @FXML
    TextField rowSizeInput, columnSizeInput;

    @FXML
    Button saveButton, loadButton;

    @FXML
    ImageView quitButton;

    private List<FieldType> fieldTypes;

    private Map customMap;
    private CustomMapPersistence persistence;

    private int rowSize;
    private int columnSize;
    private GraphicsContext graphics;
    private Renderer renderer;

    private int currentTypeIndex;

    private Field selectedField;

    MediaPlayer music;

    /**
     * Initializes the graphicsContext, the renderer, the map and its data,
     * draws the map to canvas and registers event handlers.
     */
    @FXML
    public void initialize() {
        //Music setup
        music = new MediaPlayer(ResourceLoader.loadMusic("src/main/resources/sfx/music/mad_forest.mp3"));
        music.setOnEndOfMedia(() -> music.seek(Duration.ZERO));
        music.setVolume(0.4);
        music.play();

        persistence = new CustomMapPersistence();
        fieldTypes = List.of(FieldType.values());
        customMap = new Map();
        columnSize = Map.MIN_COLUMN_NUMBER;
        rowSize = Map.MIN_ROW_NUMBER;
        customMap.setWidth(columnSize);
        customMap.setHeight(rowSize);
        settings.getStyleClass().add("settings");
        rowSizeInput.setText(String.valueOf(rowSize));
        columnSizeInput.setText(String.valueOf(columnSize));

        graphics = mapEditorCanvas.getGraphicsContext2D();
        renderer = new Renderer(customMap, graphics);
        drawGraphics();
        registerTextFieldEventHandlers();
        registerPersistenceButtons();
        registerQuitButton();
    }

    /**
     * Handles click events on canvas.
     *
     * @param event The MouseEvent
     */
    @FXML
    protected void editorCanvasClicked(MouseEvent event) {
        double row = event.getY() / mapEditorCanvas.getHeight() * rowSize;
        double column = event.getX() / mapEditorCanvas.getWidth() * columnSize;
        Node node = (Node) event.getSource();
        Stage thisStage = (Stage) node.getScene().getWindow();

        if (fieldTypeSelectingPopup == null && !customMap.nearCastle((int) column, (int) row)) {
            selectedField = customMap.getField((int) row, (int) column);
            createFieldTypeSelectingPopUp(thisStage);
        } else if (fieldTypeSelectingPopup != null) {
            hideSelectingPopUp();
        } else {
            DialogDisplayer.displayErrorDialog("Cannot place obstacles here!");
        }
    }

    /**
     * Handles mouse events on quitButton.
     */
    private void registerQuitButton() {
        quitButton.addEventHandler(MouseEvent.MOUSE_CLICKED, mouseEvent -> {
            SFXStorage.getSFX(SFXType.BUTTON_CLICK).play();

            Optional<ButtonType> result = DialogDisplayer.displayConfirmationDialog();
            if (result.isPresent() && result.get().getText().equals("Yes")) {
                Node node = (Node) mouseEvent.getSource();
                Stage thisStage = (Stage) node.getScene().getWindow();
                FXMLLoader fxmlLoader = new FXMLLoader(getClass().getResource("main-window.fxml"));
                Scene gameScene = null;
                try {
                    gameScene = new Scene(fxmlLoader.load());
                } catch (IOException e) {
                    DialogDisplayer.displayErrorDialog("The scene loading was unsuccessful!");
                }
                Objects.requireNonNull(gameScene).getStylesheets().add(Objects.requireNonNull(getClass().getResource("mmstyle.css")).toExternalForm());
                thisStage.setScene(gameScene);
                music.stop();
            }
        });
    }

    /**
     * Handles events on saveButton and loadButton.
     */
    private void registerPersistenceButtons() {
        saveButton.setOnAction(actionEvent -> {
            SFXStorage.getSFX(SFXType.BUTTON_CLICK).play();

            String path;
            CustomMapSave save = new CustomMapSave(columnSize, rowSize, customMap.getFields());
            Node node = (Node) actionEvent.getSource();
            Stage thisStage = (Stage) node.getScene().getWindow();
            File file = DialogDisplayer.displaySavingWindow(thisStage);
            if(file != null) {
                path = file.getPath();
                persistence.save(path, save);
            }
        });

        loadButton.addEventHandler(MouseEvent.MOUSE_CLICKED, mouseEvent -> {
            SFXStorage.getSFX(SFXType.BUTTON_CLICK).play();

            Node node = (Node) mouseEvent.getSource();
            Stage thisStage = (Stage) node.getScene().getWindow();
            File file = DialogDisplayer.displayLoadingWindow(thisStage);

            if(file != null) {
                String path = file.getPath();
                CustomMapSave result = persistence.load(path);
                if(result != null && !result.getFields().isEmpty()) {
                    customMap = new Map(result.getFields());
                    rowSize = customMap.getHeight();
                    columnSize = customMap.getWidth();
                    rowSizeInput.setText(String.valueOf(rowSize));
                    columnSizeInput.setText(String.valueOf(columnSize));
                    renderer.setMap(customMap);
                    drawGraphics();
                }
            }
        });
    }

    /**
     * Handles the key events on rowSizeInput and columnSizeInput.
     * The input value must be between the min and max values,
     * otherwise it sets to the closest valid size automatically.
     * The other input is modified also automatically
     * to keep the 1:2 proportion of the map.
     */
    private void registerTextFieldEventHandlers() {
        rowSizeInput.setOnKeyTyped(keyEvent -> {
            try {
                rowSize = Integer.parseInt(rowSizeInput.getText());
            } catch (NumberFormatException e) {
                rowSize = Map.MIN_ROW_NUMBER;
                rowSizeInput.setText(String.valueOf(rowSize));
            } finally {
                if(fieldTypeSelectingPopup != null) {
                    hideSelectingPopUp();
                }
                int newRowSize = Math.max(rowSize, Map.MIN_ROW_NUMBER);
                newRowSize = Math.min(newRowSize, Map.MAX_ROW_NUMBER);
                columnSize = Math.min(newRowSize*2, Map.MAX_COLUMN_NUMBER);
                modifyMapSize(newRowSize, columnSize);
                drawGraphics();
            }
        });

        columnSizeInput.setOnKeyTyped(keyEvent -> {
            try {
                columnSize = Integer.parseInt(columnSizeInput.getText());
            } catch (NumberFormatException e) {
                columnSize = Map.MIN_COLUMN_NUMBER;
                columnSizeInput.setText(String.valueOf(columnSize));
            } finally {
                if(fieldTypeSelectingPopup != null) {
                    hideSelectingPopUp();
                }
                int newColumnSize = Math.max(columnSize, Map.MIN_COLUMN_NUMBER);
                newColumnSize = Math.min(newColumnSize, Map.MAX_COLUMN_NUMBER);
                rowSize = Math.max(newColumnSize/2, Map.MIN_ROW_NUMBER);
                modifyMapSize(rowSize, newColumnSize);
                drawGraphics();
            }
        });
    }

    /**
     * Handles the click event on SelectingPopUp's labels and ImageView.
     *
     * @param prev The SelectingPopUp's 'previous' label
     * @param current The SelectingPopUp's 'current' ImageView
     * @param next The SelectingPopUp's 'next' label
     */
    private void registerPopUpLabelEventHandlers(Label prev, ImageView current, Label next) {
        customMap.changePathFinder();
        prev.addEventHandler(MouseEvent.MOUSE_CLICKED, mouseEvent -> {
            SFXStorage.getSFX(SFXType.BUTTON_CLICK).play();
            currentTypeIndex = (currentTypeIndex == 0) ? fieldTypes.size()-1 : --currentTypeIndex;
            updateCurrentTypeImageView(current);
        });

        next.addEventHandler(MouseEvent.MOUSE_CLICKED, mouseEvent -> {
            SFXStorage.getSFX(SFXType.BUTTON_CLICK).play();
            currentTypeIndex = (currentTypeIndex == fieldTypes.size()-1) ? 0 : ++currentTypeIndex;
            updateCurrentTypeImageView(current);
        });

        current.addEventHandler(MouseEvent.MOUSE_CLICKED, mouseEvent -> {
            if(!customMap.blockPath(selectedField)) {
                SFXStorage.getSFX(SFXType.BUTTON_CLICK).play();
                selectedField.setType(fieldTypes.get(currentTypeIndex));
                customMap.getPathFinder().refresh(customMap.getFields());
                hideSelectingPopUp();
                drawGraphics();
            } else {
                DialogDisplayer.displayErrorDialog("The obstacle would block the path!");
                hideSelectingPopUp();
            }
        });
    }

    /**
     * Modifies the map's row size and column size according to the given values,
     * resize the map, and modifies the input fields appropriately.
     *
     * @param rowSize The new row size
     * @param columnSize The new column size
     */
    private void modifyMapSize(int rowSize, int columnSize) {
        this.rowSize = rowSize;
        this.columnSize = columnSize;
        customMap.setWidth(columnSize);
        customMap.setHeight(rowSize);
        rowSizeInput.setText(String.valueOf(rowSize));
        columnSizeInput.setText(String.valueOf(columnSize));
    }

    /**
     * Hides the SelectingPopUp and sets the selectedField to null.
     */
    private void hideSelectingPopUp() {
        if(fieldTypeSelectingPopup != null) {
            fieldTypeSelectingPopup.hide();
            fieldTypeSelectingPopup = null;
        }
        selectedField = null;
    }

    /**
     * Creates the field type selecting popup,
     * sets its anchor points and shows it in the current stage.
     *
     * @param stage The current stage object
     */
    private void createFieldTypeSelectingPopUp(Stage stage) {
        currentTypeIndex = 0;

        fieldTypeSelectingPopup = new Popup();
        HBox selecting = new HBox();
        selecting.getStyleClass().add("towerinfo-item");
        selecting.setAlignment(Pos.CENTER);

        Label previousType = new Label("◀");
        Label nextType = new Label("▶");
        ImageView currentType = new ImageView();
        currentType.setFitWidth(100);
        currentType.setFitHeight(100);
        updateCurrentTypeImageView(currentType);
        previousType.getStyleClass().add("selecting-label");
        nextType.getStyleClass().add("selecting-label");
        currentType.getStyleClass().add("selecting-imageview");
        registerPopUpLabelEventHandlers(previousType, currentType, nextType);

        selecting.getChildren().addAll(previousType, currentType, nextType);

        fieldTypeSelectingPopup.getContent().add(selecting);
        fieldTypeSelectingPopup.show(stage);
        setPopupAnchors(fieldTypeSelectingPopup, stage);
    }

    /**
     * Renders the graphics of the map editor.
     */
    private void drawGraphics() {
        graphics.clearRect(0, 0, mapEditorCanvas.getWidth(), mapEditorCanvas.getHeight());
        renderer.setHeightRatio(mapEditorCanvas.getHeight() / customMap.getHeight());
        renderer.setWidthRatio(mapEditorCanvas.getWidth() / customMap.getWidth());
        customMap.setPlayersCastleToDefault();
        renderer.drawFields();
        renderer.drawCastles();
        renderer.drawGrid(mapEditorCanvas);
    }

    /**
     * Updates the current field's image according to the selected field type index.
     *
     * @param currentField The current field's ImageView object
     */
    private void updateCurrentTypeImageView(ImageView currentField) {
        currentField.setImage(ImageStorage.getImage(fieldTypes.get(currentTypeIndex).getImagePath()));
    }

    /**
     * Sets the popup's position above the selected field.
     *
     * @param popup The popup instance
     * @param stage The window
     */
    private void setPopupAnchors(Popup popup, Stage stage) {

        if (popup != null) {
            int row = (int)selectedField.getPosition().getY();
            int column = (int)selectedField.getPosition().getX();
            double canvasPosX = stage.getX() + mapEditorCanvas.getScene().getX() + mapEditorCanvas.getBoundsInParent().getMinX();
            double canvasPosY = stage.getY() + mapEditorCanvas.getScene().getY() + mapEditorCanvas.getParent().getBoundsInParent().getMinY();

            double widthRatio = mapEditorCanvas.getWidth() / columnSize;
            double heightRatio = mapEditorCanvas.getHeight() / rowSize;

            popup.setAnchorX(canvasPosX + column*widthRatio - popup.getWidth()/2 + widthRatio/2);
            popup.setAnchorY(canvasPosY + row*heightRatio - popup.getHeight());
        }
    }

    /**
     * Moves the current popup's position according to window location.
     *
     * @param stage The window
     */
    public void movePopupWithWindow(Stage stage) {
        if(fieldTypeSelectingPopup != null) {
            this.setPopupAnchors(fieldTypeSelectingPopup, stage);
            handlePopupVisibility(fieldTypeSelectingPopup, stage);
        }
    }

    /**
     * Hides or shows the popup according to the window location.
     *
     * @param popup The Popup instance
     * @param stage The window
     */
    private void handlePopupVisibility(Popup popup, Stage stage) {
        Rectangle2D screenBounds = Screen.getPrimary().getVisualBounds();
        if(popup.getAnchorX() + popup.getWidth() >= screenBounds.getMaxX()
                || popup.getAnchorY() + popup.getHeight() >= screenBounds.getMaxY()
                || popup.getAnchorX() <= screenBounds.getMinX()
                || popup.getAnchorY() <= screenBounds.getMinY()) {
            popup.hide();
        } else {
            popup.show(stage);
        }
    }
}