package bitsplease.mummymayhem.model;

import bitsplease.mummymayhem.model.buildings.Barrack;
import bitsplease.mummymayhem.model.buildings.Building;
import bitsplease.mummymayhem.model.buildings.Tower;
import bitsplease.mummymayhem.model.map.Field;
import bitsplease.mummymayhem.model.map.FieldType;
import bitsplease.mummymayhem.model.map.Map;
import bitsplease.mummymayhem.model.units.*;
import bitsplease.mummymayhem.model.utils.ClimberPathFinder;
import bitsplease.mummymayhem.model.utils.PathFinder;
import bitsplease.mummymayhem.model.utils.Sprite;
import javafx.geometry.Point2D;

import java.util.*;

public class GameModel {

    public static final List<Unit> units = List.of(
            new Demolisher(new Point2D(0,0), "src/main/resources/assets/units/demolisher/"),
            new DuneClimber(new Point2D(0,0), "src/main/resources/assets/units/duneClimber/"),
            new Fighter(new Point2D(0,0), "src/main/resources/assets/units/fighter/"),
            new Runner(new Point2D(0,0), "src/main/resources/assets/units/runner/"),
            new Tank(new Point2D(0,0), "src/main/resources/assets/units/tank/"));

    private static final int DEFAULT_WIDTH = 30;
    private static final int DEFAULT_HEIGHT = 15;
    private static final int BARRACK_LIMIT = 3;

    private final Map map;
    private final Player playerA;
    private final Player playerB;

    private PhaseType phase;
    private Sprite selectedBuilding;

    private final List<Point2D> playerAFieldsPositions;
    private final List<Point2D> playerBFieldsPositions;

    public GameModel() {
        this.map = new Map(DEFAULT_WIDTH, DEFAULT_HEIGHT);
        playerA = new Player(PlayerType.PLAYER_A, map.getPlayerACastlePosition());
        playerB = new Player(PlayerType.PLAYER_B, map.getPlayerBCastlePosition());
        addCastlesToFields();
        phase = PhaseType.PLAYER_A_BUILD;
        PhaseType.reset();
        selectedBuilding = null;

        playerAFieldsPositions = map.getPlayerAFieldsPositions();
        playerBFieldsPositions = map.getPlayerBFieldsPositions();
    }

    public GameModel(Map map) {
        this.map = map;
        playerA = new Player(PlayerType.PLAYER_A, map.getPlayerACastlePosition());
        playerB = new Player(PlayerType.PLAYER_B, map.getPlayerBCastlePosition());
        addCastlesToFields();
        phase = PhaseType.PLAYER_A_BUILD;
        PhaseType.reset();
        selectedBuilding = null;
        playerAFieldsPositions = map.getPlayerAFieldsPositions();
        playerBFieldsPositions = map.getPlayerBFieldsPositions();
    }

    /**
     * Calls PhaseType's nextPhase method.
     */
    public void nextPhase() {
        phase = PhaseType.nextPhase();
    }

    public Map getMap() {
        return map;
    }

    public Player getPlayerA() {
        return playerA;
    }

    public Player getPlayerB() {
        return playerB;
    }

    public Player getActualPlayer() {
        if(phase == PhaseType.PLAYER_A_BUILD) {
            return playerA;
        } else {
            return playerB;
        }
    }

    public PhaseType getPhase() {
        return phase;
    }

    public Sprite getSelectedBuilding() {
        return selectedBuilding;
    }

    /**
     * Sets the 'selectedBuilding' field to the given building.
     *
     * @param building The building to focus
     */
    public void selectBuilding (Sprite building) {
        selectedBuilding = building;
    }

    /**
     * Returns the winner of the game.
     *
     * @return The winner's Player object or null if the game was draw
     */
    public PlayerType getWinner() {
        if (!getPlayerA().getCastle().isAlive() && !getPlayerB().getCastle().isAlive()) {
            return null;
        }
        else if (!getPlayerA().getCastle().isAlive()) {
            return PlayerType.PLAYER_B;
        }
        else {
            return PlayerType.PLAYER_A;
        }
    }

    /**
     * Checks if at least one of the castles has 0 hp.
     *
     * @return true if at least one of the castles has 0 hp, false otherwise
     */
    public boolean gameEnded() {
        return !getPlayerA().getCastle().isAlive() || !getPlayerB().getCastle().isAlive();
    }

    /**
     * Checks if current player already has 3 barracks.
     *
     * @return true if current player already has 3 barracks, false otherwise
     */
    public boolean playerReachedBarrackLimit(){
        return getActualPlayer().getBarracks().size() >= BARRACK_LIMIT;
    }

    /**
     * Checks if there is a barrack on the given position.
     *
     * @param column The column number
     * @param row The row number
     * @return true if barrack found, false otherwise
     */
    public boolean isBarrackPosition(int column, int row) {
        List<Barrack> barracks = new ArrayList<>();
        barracks.addAll(playerA.getBarracks());
        barracks.addAll(playerB.getBarracks());

        for(Barrack b: barracks) {
            if((int)b.getPosition().getX() == column && (int)b.getPosition().getY() == row) {
                return true;
            }
        }
        return false;
    }

    /**
     * Checks if there is a tower on the given position.
     *
     * @param column The column number
     * @param row The row number
     * @return true if tower found, false otherwise
     */
    public boolean isTowerPosition(int column, int row) {
        List<Tower> towers = new ArrayList<>();
        towers.addAll(playerA.getTowers());
        towers.addAll(playerB.getTowers());

        for(Tower tower: towers) {
            if((int)tower.getPosition().getX() == column && (int)tower.getPosition().getY() == row) {
                return true;
            }
        }
        return false;
    }

    /**
     * Checks the position's surroundings.
     * The position is valid if it's on the player's own half
     * and there is no barrack among its neighbours.
     *
     * @param column The column number
     * @param row The row number
     * @return true if it's a valid position, false otherwise
     */
    public boolean isValidTowerPosition(int column, int row) {

        Point2D position = new Point2D(column, row);
        List<Point2D> directions = List.of(
                new Point2D(1, 0),
                new Point2D(0, 1),
                new Point2D(-1, 0),
                new Point2D(0, -1));

        List<Point2D> existingPositions = map.getFields().stream()
                .flatMap(Collection::stream)
                .map(Sprite::getPosition)
                .toList();

        if((this.getActualPlayer().equals(playerA) && playerAFieldsPositions.contains(position))
            || (this.getActualPlayer().equals(playerB) && playerBFieldsPositions.contains(position))) {

            for (Point2D direction : directions) {
                Point2D neighbor = new Point2D(column + direction.getX(), row + direction.getY());

                if (existingPositions.contains(neighbor)
                    && isBarrackPosition((int)neighbor.getX(), (int)neighbor.getY())) {
                        return false;
                }
            }
            return true;
        }
        return false;
    }

    /**
     * Checks the position's surroundings.
     * The position is valid if it's on the player's own half
     * and the position is close enough to the corresponding castle
     * and there is no tower among its neighbours.
     *
     * @param column The column number
     * @param row The row number
     * @return true if it's a valid position, false otherwise
     */
    public boolean isValidBarrackPosition(int column, int row) {

        Point2D position = new Point2D(column, row);
        List<Point2D> directions = List.of(
                new Point2D(1, 0),
                new Point2D(0, 1),
                new Point2D(-1, 0),
                new Point2D(0, -1));

        List<Point2D> existingPositions = map.getFields().stream()
                .flatMap(Collection::stream)
                .map(Sprite::getPosition)
                .toList();

        List<Point2D> possibleBarrackPositions;

        if((this.getActualPlayer().equals(playerA) && playerAFieldsPositions.contains(position))
            || (this.getActualPlayer().equals(playerB) && playerBFieldsPositions.contains(position))) {

            if (this.getActualPlayer().equals(playerA))
                possibleBarrackPositions = this.getMap().getPlayerAPossibleBarrackPositions();
            else
                possibleBarrackPositions = this.getMap().getPlayerBPossibleBarrackPositions();

            if(!possibleBarrackPositions.contains(position))
                return false;

            for (Point2D direction : directions) {
                Point2D neighbor = new Point2D(column + direction.getX(), row + direction.getY());

                if (existingPositions.contains(neighbor)
                        && isTowerPosition((int)neighbor.getX(), (int)neighbor.getY())) {
                        return false;
                }
            }
            return true;
        }
        return false;
    }

    /**
     * Counts the player's units with the same class and start position.
     *
     * @param unit The unit sample
     * @param column The column number
     * @param row The row number
     * @return The size of the list of units with the same type
     */
    public int getUnitBoughtNumber(Unit unit, int column, int row) {

        return this.getUnitsWithSameType(unit, new Point2D(column, row)).size();
    }

    /**
     * Method which implements the unit purchase,
     * and sets the purchased unit's default path.
     *
     * @param unit The sample of the unit to purchase
     * @param position The start position of the unit
     * @return The bought Unit instance
     */
    public Unit playerBuyUnit(Unit unit, Point2D position) {

        Unit newUnit;
        Point2D destination;

        if (getActualPlayer().equals(playerA)) {
            destination = playerB.getCastle().getPosition();
        }
        else {
            destination = playerA.getCastle().getPosition();
        }

        if(unit instanceof Fighter) {
            newUnit = new Fighter(unit);
        }
        else if(unit instanceof Demolisher) {
            newUnit = new Demolisher(unit);
            destination = getClosestDemolisherDestination(position);
        }
        else if(unit instanceof DuneClimber) {
            newUnit = new DuneClimber(unit);
        }
        else if(unit instanceof Runner) {
            newUnit = new Runner(unit);
        }
        else {
            newUnit = new Tank(unit);
        }

        setDefaultPath(newUnit, position, destination);
        if(getActualPlayer().buyUnit(newUnit)) {
            return newUnit;
        }
        return null;
    }

    /**
     * Adds or deletes the given position from unit's path,
     * and applies changes for all units (which has the same class
     * and same start position) possessed by the appropriate player.
     *
     * @param unitSample The unit sample
     * @param unitPosition The start position of the unit
     * @param column The column number of the selected position
     * @param row The row number of the selected position
     */
    public void modifyUnitPath(Unit unitSample, Point2D unitPosition, int column, int row) {

        Unit unit;
        Point2D position = new Point2D(column, row);
        boolean addCheckpoint = true;
        Field selectedField = map.getField((int)position.getY(), (int)position.getX());
        List<Unit> unitsWithSameType = this.getUnitsWithSameType(unitSample, unitPosition);

        if (unitsWithSameType.size() == 0) {
            return;
        }
        else {
            unit = unitsWithSameType.get(0);
        }

        if(unit instanceof Demolisher) {

            if(!isValidDemolisherCheckpoint(unit, selectedField)) {
                return;
            }

            //if player selected a valid tower, set a new default path
            //in this case adding checkpoint isn't needed
            //previous checkpoints will be applied to this default path
            if (selectedField.hasTower()) {
                unit.calculateDefaultPath(map, position);
                addCheckpoint = false;
            }
        }

        if(unit instanceof DuneClimber) {
            if(!isValidClimberCheckpoint(unit, selectedField)) {
                return;
            }
        }

        if(!(unit instanceof Demolisher) && !(unit instanceof DuneClimber)) {
            if(!isValidRegularUnitCheckpoint(selectedField)) {
                return;
            }
        }

        if (addCheckpoint) {
            if (!unit.getCheckPoints().contains(position)) {
                unit.addCheckPoint(position);
            }
            else {
                unit.deleteCheckPoint(position);
            }
        }

        List<Point2D> newPositions = unit.getUpdatedPath(map);
        List<Field> newPath = this.getFieldsFromPoints(newPositions);

        for (Unit u : unitsWithSameType) {
            u.setPath(newPath);
            u.setCheckPoints(unit.getCheckPoints());
        }
    }


    /**
     * Method which needs to be called after building.
     */
    public void refreshUnitsPath(int column, int row) {
        refreshPlayerUnitsPath(column, row, playerA);
        refreshPlayerUnitsPath(column, row, playerB);
    }

    /**
     * Initializes all units' departing priority.
     */
    public void initializeUnitPriorities() {
        initializePlayerUnitPriority(playerA);
        initializePlayerUnitPriority(playerB);
    }

    /**
     * Initializes a player's units' departing priority.
     * @param player The Player object
     */
    private void initializePlayerUnitPriority (Player player){
        ArrayList<Building> buildings = new ArrayList<>();

        buildings.add(player.getCastle());
        buildings.addAll(player.getBarracks());

        for(Building b: buildings) {
            int nextPriority = 0;
            for (Unit u: player.getUnits()) {
                if(u.getPosition().equals(b.getPosition())) {
                    u.setPriority(nextPriority);
                    nextPriority++;
                }
            }
        }
    }

    /**
     * Deletes checkpoints which became unreachable,
     * and modifies player's units' paths if they collide with a building.
     *
     * @param column The column number
     * @param row The row number
     * @param player The player whose units are considered
     */
    private void refreshPlayerUnitsPath(int column, int row, Player player) {

        Point2D position = new Point2D(column, row);

        for (Unit unit : player.getUnits()) {

            List<Point2D> path = unit.getPath().stream().map(Sprite::getPosition).toList();
            if(path.contains(position)) {

                unit.updateCheckpoints(map);
                List<Point2D> newPositions = unit.getUpdatedPath(map);
                List<Field> newPath = this.getFieldsFromPoints(newPositions);
                unit.setPath(newPath);
            }
        }
    }

    /**
     * Adds the players' castles to the map's fields.
     */
    private void addCastlesToFields() {
        double castleAX = playerA.getCastle().getPosition().getX();
        double castleAY = playerA.getCastle().getPosition().getY();
        double castleBX = playerB.getCastle().getPosition().getX();
        double castleBY = playerB.getCastle().getPosition().getY();

        map.getFields().get((int)castleAX).get((int)castleAY).addBuilding(playerA.getCastle());
        map.getFields().get((int)castleBX).get((int)castleBY).addBuilding(playerB.getCastle());
    }

    /**
     * Converts a list of points to a list of fields.
     *
     * @param positions The list of points
     * @return The list of fields
     */
    private List<Field> getFieldsFromPoints(List<Point2D> positions) {

        List<Field> newPath = new ArrayList<>();

        for (Point2D p : positions) {
            newPath.add(map.getFields().get((int)p.getX()).get((int)p.getY()));
        }

        return newPath;
    }

    /**
     * Returns a list of current player's units which class is the same as the unitSample's
     * and also have the given start position.
     *
     * @param unitSample The unit sample
     * @param unitPosition The start position
     * @return The list of units of the same type
     */
    private List<Unit> getUnitsWithSameType(Unit unitSample, Point2D unitPosition) {
        List<Unit> unitsWithSameType;

        unitsWithSameType = getActualPlayer().getUnits().stream()
                .filter(u -> u.getClass().equals(unitSample.getClass()))
                .filter(u -> u.getPosition().equals(unitPosition))
                .toList();

        return unitsWithSameType;
    }


    /**
     * Checks if the selected position is valid or not.
     * It's valid if the selected field has an alive reachable enemy tower,
     * or if it's an open reachable field.
     *
     * @param unit The Demolisher instance
     * @param selectedField The Field object of the selected position
     * @return true if the selected position is valid, false otherwise
     */
    private boolean isValidDemolisherCheckpoint(Unit unit, Field selectedField) {

        Point2D position = selectedField.getPosition();

        if(selectedField.hasTower()) {

            List<Point2D> ownTowers = getActualPlayer().getTowers().stream()
                    .map(Sprite::getPosition)
                    .toList();

            if(!ownTowers.contains(position) && !selectedField.getTower().isDemolished()){
                PathFinder pathFinder = new PathFinder(map.getWidth(), map.getHeight(),
                        unit.getPosition(), position, map.getFields());

                return pathFinder.isTraversable();
            }
            else {
                return false;
            }
        }
        else if(selectedField.isOccupied()) {
            return false;
        }
        else return map.getPathFinder().isTraversable(position);
    }

    /**
     * Checks if the selected position is valid or not.
     * It's valid if the selected field is a reachable dune,
     * or if it's an open reachable field.
     *
     * @param unit The DuneClimber instance
     * @param selectedField The Field object of the selected position
     * @return true if the selected position is valid, false otherwise
     */
    private boolean isValidClimberCheckpoint(Unit unit, Field selectedField) {

        Point2D position = selectedField.getPosition();

        if(selectedField.isOccupied()) {
            if (selectedField.getType() == FieldType.DUNE) {
                PathFinder pathFinder = new ClimberPathFinder(map.getWidth(), map.getHeight(),
                        unit.getPosition(), position, map.getFields());

                return pathFinder.isTraversable();
            }
            else {
                return false;
            }
        }
        else {
            PathFinder pathFinder = new ClimberPathFinder(map.getWidth(), map.getHeight(),
                    unit.getPosition(), position, map.getFields());

            return pathFinder.isTraversable();
        }
    }

    /**
     * Checks if the selected position is valid or not.
     * It's valid if the selected field is a reachable open field.
     *
     * @param selectedField The Field object of the selected position
     * @return true if the selected position is valid, false otherwise
     */
    private boolean isValidRegularUnitCheckpoint(Field selectedField) {

        Point2D position = selectedField.getPosition();

        if(!selectedField.isOccupied()) {
            return map.getPathFinder().isTraversable(position);
        }
        else {
            return false;
        }
    }

    /**
     * Returns the closest alive enemy tower's position from the given position,
     * or the enemy castle's position if tower not found.
     *
     * @param position The Demolisher's start position
     * @return The destination
     */
    private Point2D getClosestDemolisherDestination(Point2D position) {

        Player otherPlayer;
        if(getActualPlayer().equals(playerA)) {
            otherPlayer = playerB;
        }
        else {
            otherPlayer = playerA;
        }

        List<Tower> enemyTowers = new ArrayList<>(otherPlayer.getTowers());
        enemyTowers.sort(Comparator.comparing(t -> t.getPosition().distance(position)));
        for (Tower t : enemyTowers) {
            PathFinder pathFinder = new PathFinder(map.getWidth(), map.getHeight(),
                    position, t.getPosition(), map.getFields());

            if(pathFinder.isTraversable() && !t.isDemolished()) {
                return t.getPosition();
            }
        }

        return otherPlayer.getCastle().getPosition();
    }

    /**
     * Sets the path of the given unit.
     * If the current player has a similar unit then the new unit inherits
     * its path and checkpoints. Otherwise, a new default path is set.
     *
     * @param newUnit The unit to handle
     * @param position The unit's start position
     * @param destination The unit's destination
     */
    private void setDefaultPath(Unit newUnit, Point2D position, Point2D destination) {

        newUnit.setPosition(position);

        Unit similarUnit = null;
        List<Unit> unitsWithSameType = this.getUnitsWithSameType(newUnit, position);
        if (unitsWithSameType.size() > 0)
            similarUnit = unitsWithSameType.get(0);

        if(similarUnit != null) {
            newUnit.setPath(similarUnit.getPath());
            newUnit.setCheckPoints(similarUnit.getCheckPoints());
        }
        else {
            newUnit.calculateDefaultPath(map, destination);
        }
    }
}
