package bitsplease.mummymayhem.model;

import bitsplease.mummymayhem.model.buildings.Barrack;
import bitsplease.mummymayhem.model.buildings.Castle;
import bitsplease.mummymayhem.model.buildings.Tower;
import bitsplease.mummymayhem.model.units.Unit;
import javafx.geometry.Point2D;

import java.util.ArrayList;
import java.util.List;

public class Player {

    private static final int INITIAL_GOLD_AMOUNT = 1000;
    private static final int DEMOLISH_PRICE = 30;

    private final Castle castle;
    private final List<Unit> units;
    private final List<Tower> towers;
    private final List<Barrack> barracks;
    private int gold;
    int trainedUnits = 0;

    public Player(PlayerType type, Point2D castlePosition) {
        this.castle = new Castle(castlePosition, type.getImagePath());
        this.gold = INITIAL_GOLD_AMOUNT;
        units = new ArrayList<>();
        towers = new ArrayList<>();
        barracks = new ArrayList<>();
    }

    public int getGold() {
        return gold;
    }

    public void setGold(int gold) {
        this.gold = gold;
    }

    public Castle getCastle() {
        return castle;
    }

    public List<Unit> getUnits() {
        return units;
    }

    public List<Tower> getTowers() {
        return towers;
    }

    public List<Barrack> getBarracks() {
        return barracks;
    }


    /**
     * Increases the player's gold amount based on the unit's price
     * @param u The unit whose price is checked
     */
    public void addGold(Unit u) {
        gold += u.getPrice() * 0.3;
    }

    /**
     * Increases the player's gold amount based on how many units they have trained so far
     */
    public void addNextRoundGold() {
        gold += 500;
        gold += trainedUnits * 10;
    }

    /**
     * Subtracts the unit's price from player's gold,
     * and adds unit to player's unit collection.
     * Does nothing if player hasn't enough gold.
     *
     * @param unit The unit to buy
     * @return true if the purchase was successful, false otherwise
     */
    public boolean buyUnit(Unit unit) {
        if(gold < unit.getPrice()) { return false; }
        gold -= unit.getPrice();
        units.add(unit);
        trainedUnits++;
        return true;
    }

    /**
     * Deletes a unit from player's unit collection if found a match.
     * Does nothing if there isn't a corresponding unit in the collection.
     *
     * @param unit The unit to delete
     * @param position The position of the unit
     * @return true if deletion was successful, false otherwise
     */
    public boolean sellUnit(Unit unit, Point2D position) {
        for(Unit u: units) {
            if(u.getClass() == unit.getClass() && u.getPosition().getX() == position.getX() && u.getPosition().getY() == position.getY()) {
                units.remove(u);
                gold += u.getPrice();
                trainedUnits--;
                return true;
            }
        }
        return false;
    }

    /**
     * Subtracts the tower's price from player's gold,
     * and adds tower to player's tower collection.
     * Does nothing if player hasn't enough gold.
     *
     * @param tower The tower to buy
     * @return true if the purchase was successful, false otherwise
     */
    public boolean buyTower(Tower tower) {
        if(gold < tower.getPrice()) { return false; }
        gold -= tower.getPrice();
        towers.add(tower);
        return true;
    }

    /**
     * Subtracts the upgrade price from player's gold,
     * and upgrades the given tower.
     * Does nothing if player hasn't enough money.
     *
     * @param tower The tower to upgrade
     * @return true if the upgrade was successful, false otherwise
     */
    public boolean upgradeTower(Tower tower) {
        if(gold < tower.getPrice()*0.5) { return false; }
        tower.upgrade();
        gold -= tower.getPrice()*0.5;
        return true;
    }

    /**
     * Subtracts the barrack's price from player's gold,
     * and adds barrack to player's barrack collection.
     * Does nothing if player hasn't enough gold.
     *
     * @param barrack The barrack to buy
     * @return true if the purchase was successful, false otherwise
     */
    public boolean buyBarrack(Barrack barrack) {
        if(gold < Barrack.getPrice()) { return false; }
        gold -= Barrack.getPrice();
        barracks.add(barrack);
        return true;
    }

    /**
     * Deletes the tower from player's tower collection.
     * Adds the half price of the tower to player's gold if it wasn't destroyed before.
     *
     * @param tower The tower to demolish
     */
    public boolean removeTower(Tower tower) {
        if(!tower.isDemolished()) {
            towers.remove(tower);
            gold += tower.getPrice()*0.5;
            return true;
        }
        else if(gold >= DEMOLISH_PRICE) {
            towers.remove(tower);
            gold -= DEMOLISH_PRICE;
            return true;
        }
        return false;
    }

    /**
     * Checks if player possesses the given tower.
     *
     * @param tower The tower to search for
     * @return true if found, false otherwise
     */
    public boolean searchTower(Tower tower) {
        return towers.contains(tower);
    }

    /**
     * Checks if player has a barrack on the given position.
     *
     * @param column The column number
     * @param row The row number
     * @return true if barrack found, false otherwise
     */
    public boolean hasBarrackOnPosition(int column, int row) {
        for(Barrack b: barracks) {
            if((int)b.getPosition().getX() == column && (int)b.getPosition().getY() == row) {
                return true;
            }
        }
        return false;
    }
}
