package bitsplease.mummymayhem.model.buildings;

import bitsplease.mummymayhem.model.utils.Sprite;
import bitsplease.mummymayhem.model.units.Unit;
import javafx.geometry.Point2D;

import java.util.List;

public class Tower extends Sprite {
    public static final long ATTACK_COOLDOWN_TIME = 500_000_000;
    private static final String PROJECTILE_FILE_PATH = "src/main/resources/assets/buildings/tower-projectile.png";
    private static final String RUINED_TOWER_FILE_PATH = "src/main/resources/assets/buildings/ruined-tower.png";
    public static final int HP_INCREASE_PER_UPGRADE = 10;
    public static final int ATTACK_INCREASE_PER_UPGRADE = 2;
    public static final int RANGE_INCREASE_PER_UPGRADE = 1;
    public static final int MAX_LEVEL = 3;

    private final int initialHp;

    private int attack;
    private int hp;
    private int range;
    private int level;
    private boolean isDemolished;
    private final int price;
    private final TowerType type;
    private final UpgradeType upgradeType;

    private Unit target;
    private long attackCooldown = 0;
    private long lastCooldownUpdate = 0;

    private long lastProjectileUpdate = 0;
    private Point2D projectilePosition;

    public Tower(Point2D position, TowerType type) {
        super(position, type.getImagePath());
        this.attack = type.getAttack();
        this.hp = type.getHp();
        this.range = type.getRange();
        this.price = type.getPrice();
        this.level = 1;
        this.isDemolished = false;
        this.type = type;
        this.upgradeType = type.getUpgradeType();

        this.initialHp = this.hp;

        this.projectilePosition = new Point2D(position.getX(), position.getY());
        this.target = null;
    }

    public Tower(TowerType type) {
        super(new Point2D(0,0), type.getImagePath());
        this.attack = type.getAttack();
        this.hp = type.getHp();
        this.range = type.getRange();
        this.price = type.getPrice();
        this.level = 1;
        this.isDemolished = false;
        this.type = type;
        this.upgradeType = type.getUpgradeType();

        this.initialHp = this.hp;

        this.target = null;
    }


    @Override
    public void setPosition(Point2D position) {
        this.position = position;
        resetProjectilePosition();
    }

    public int getAttack() {
        return attack;
    }

    public void setAttack(int attack) {
        this.attack = attack;
    }

    public void setHp(int hp) {
        this.hp = hp;
    }

    public void setRange(int range) {
        this.range = range;
    }

    public int getHp() {
        return hp;
    }

    public int getRange() {
        return range;
    }

    public int getPrice() {
        return price;
    }

    public int getLevel() {
        return level;
    }

    public TowerType getType() {
        return type;
    }

    public UpgradeType getUpgradeType() { return upgradeType; }

    /**
     * Returns if tower is demolished.
     *
     * @return true if tower is demolished, false otherwise
     */
    public boolean isDemolished() {
        return isDemolished;
    }

    /**
     * Sets the tower state. Modifies its image path if demolished.
     *
     * @param isDemolished true if demolished, false otherwise
     */
    public void setDemolished(boolean isDemolished) {
        this.isDemolished = isDemolished;
        if(isDemolished) {
            this.imagePath = Tower.RUINED_TOWER_FILE_PATH;
            this.resetProjectilePosition();
            this.target = null;
        }
    }

    public long getAttackCooldown() {
        return attackCooldown;
    }

    public Unit getTarget() {
        return target;
    }

    public long getLastCooldownUpdate() {
        return lastCooldownUpdate;
    }

    public void setLastCooldownUpdate(long lastCooldownUpdate) {
        this.lastCooldownUpdate = lastCooldownUpdate;
    }

    public long getLastProjectileUpdate() {
        return lastProjectileUpdate;
    }

    public void setLastProjectileUpdate(long lastProjectileUpdate) {
        this.lastProjectileUpdate = lastProjectileUpdate;
    }

    public Point2D getProjectilePosition() {
        return projectilePosition;
    }

    public void setProjectilePosition(Point2D projectilePosition) {
        this.projectilePosition = projectilePosition;
    }

    public String getProjectileFilePath() {
        return PROJECTILE_FILE_PATH;
    }

    public int getInitialHp() {
        return initialHp;
    }

    /**
     * Resets the projectile to the tower's position
     */
    public void resetProjectilePosition() {
        this.projectilePosition = new Point2D(position.getX(), position.getY());
    }

    /**
     * Lowers the attack cooldown by a certain value
     * @param l The value to lower the cooldown by
     */
    public void lowerAttackCooldown(long l) {
        if(this.attackCooldown - l <= 0) {
            this.attackCooldown = 0;
        }
        else {
            this.attackCooldown -= l;
        }
    }

    /**
     * Attacks a unit, decreasing its health by the tower's attack value
     *
     * @param unit The target unit
     */
    public void attack(Unit unit) {
        unit.takeDamage(this.attack);

        this.attackCooldown = ATTACK_COOLDOWN_TIME;
        resetProjectilePosition();

    }

    /**
     * Decreases the tower's hp by a given value,
     * or sets hp to 0 if the value is higher than current hp.
     *
     * @param damage The value of the incoming damage
     */
    public void takeDamage(int damage) {
        if(hp <= damage) {
            hp = 0;
            this.setDemolished(true);
        }
        else {
            hp -= damage;
        }
    }

    /**
     * Upgrades the tower,
     * increasing its health, hp, attack and range
     */
    public void upgrade() {
        if(level >= MAX_LEVEL || isDemolished) {
            return;
        }

        level += 1;
        switch (upgradeType) {
            case HP -> hp += HP_INCREASE_PER_UPGRADE;
            case ATTACK -> attack += ATTACK_INCREASE_PER_UPGRADE;
            case RANGE -> range += RANGE_INCREASE_PER_UPGRADE;
        }
    }

    /**
     * Targets the unit that's closest to their destination within the range of the tower
     * from the given List. If the projectile is currently on its way towards a unit it does nothing.
     * If the currently targeted unit goes out of range, dies or reaches its destination it selects a new target.
     * @param units The list of units to select the target from
     */
    public void targetUnit(List<Unit> units) {
        if(!projectilePosition.equals(position)) {
            return;
        }

        double towerX = position.getX();
        double towerY = position.getY();

        if(target != null
                && (Math.abs(towerX - target.getPosition().getX()) > range
                    || Math.abs(towerY - target.getPosition().getY()) > range
                    || target.getDistanceFromDestination() <= 0
                    || target.getHp() == 0)) {
            target = null;
        }

        double minDistance;

        for(Unit u: units) {
            if(target == null) {
                minDistance = u.getDistanceFromDestination();
            }
            else {
                minDistance = target.getDistanceFromDestination();
            }

            double unitX = u.getPosition().getX();
            double unitY = u.getPosition().getY();

            if(Math.abs(towerX - unitX) <= range && Math.abs(towerY - unitY) <= range && minDistance >= u.getDistanceFromDestination()) {
                if(target == null || !target.equals(u)) {
                    target = u;
                }
            }
        }
    }
}
