package bitsplease.mummymayhem.model.map;

import bitsplease.mummymayhem.model.buildings.Building;
import bitsplease.mummymayhem.model.buildings.Tower;
import bitsplease.mummymayhem.model.utils.PathFinder;
import bitsplease.mummymayhem.model.utils.Sprite;
import javafx.geometry.Point2D;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

public class Map {

    public static final int MIN_ROW_NUMBER = 4;
    public static final int MIN_COLUMN_NUMBER = 8;
    public static final int MAX_ROW_NUMBER = 25;
    public static final int MAX_COLUMN_NUMBER = 50;

    private static final Random random = new Random();

    private int width;
    private int height;
    private Point2D playerACastlePosition;
    private Point2D playerBCastlePosition;
    private PathFinder pathFinder;

    private List<List<Field>> fields;
    private List<Point2D> coordinatesNearCastles;

    public Map() {
        this.width = 0;
        this.height = 0;
        this.playerACastlePosition = new Point2D(0, 0);
        this.playerBCastlePosition = new Point2D(0, 0);
        initEmptyMap();
        this.pathFinder = new PathFinder(width, height, playerACastlePosition, playerBCastlePosition, fields);
    }

    public Map(int width, int height) {
        this.width = width;
        this.height = height;
        setPlayersCastleToDefault();
        initEmptyMap();
        generateFields();

        this.pathFinder = new PathFinder(width, height, playerACastlePosition, playerBCastlePosition, fields);

        while(!pathFinder.isTraversable()) {
            generateFields();
            pathFinder.refresh(fields);
        }
    }

    public Map(List<List<Field>> fields) {
        this.fields = fields;
        this.width = fields.size();
        this.height = fields.get(0).size();

        setPlayersCastleToDefault();

        this.pathFinder = new PathFinder(width, height, playerACastlePosition, playerBCastlePosition, fields);
    }

    public void setWidth(int width) {
        this.width = width >= MIN_COLUMN_NUMBER ? width : 0;
        initEmptyMap();
    }

    public void setHeight(int height) {
        this.height = height >= MIN_ROW_NUMBER ? height : 0;
        initEmptyMap();
    }

    public void changePathFinder() {
        this.pathFinder = new PathFinder(width, height, playerACastlePosition, playerBCastlePosition, fields);
    }

    public Field getField(int row, int column) { return fields.get(column).get(row); }

    public void setField(int x, int y, FieldType type) {
        fields.get(x).set(y, new Field(new Point2D(x,y), type));
    }

    public int getWidth() {
        return width;
    }

    public int getHeight() {
        return height;
    }

    public PathFinder getPathFinder() {
        return pathFinder;
    }

    public Point2D getPlayerACastlePosition() {
        return playerACastlePosition;
    }

    public void setPlayersCastleToDefault() {
        this.playerACastlePosition = new Point2D(0, (double)(height / 2) - 1);

        if (height % 2 == 0) {
            this.playerBCastlePosition = new Point2D(width - 1, (double)height / 2);
        }
        else {
            this.playerBCastlePosition = new Point2D(width - 1, (double)(height / 2) + 1);
        }
        generateCoordinatesNearCastles();
    }

    public Point2D getPlayerBCastlePosition() {
        return playerBCastlePosition;
    }

    public List<List<Field>> getFields() {
        return fields;
    }

    /**
     * Checks if a building were to be placed, would it block the path between the castles,
     * or if it's near a castle.
     *
     * @param building The building to be checked
     * @return If the building will block the path or not, or the building is near a castle
     */
    public boolean blockPath(Sprite building) {
        return nearCastle(building.getPosition()) || pathFinder.blockPath(building.getPosition());
    }

    /**
     * Initializes the ArrayLists containing the fields
     */
    private void initEmptyMap() {
        this.fields = new ArrayList<>(width);

        for(int i = 0; i < width; i++) {
            fields.add(new ArrayList<>(height));
        }

        for (int i = 0; i < width; i++) {
            for (int j = 0; j < height; j++) {
                fields.get(i).add(new Field(new Point2D(i, j), FieldType.OPEN));
            }
        }
    }

    /**
     * Generates a random map.
     * It sets all the fields to open,
     * then randomly selects 20% of the fields to be occupied,
     * The fields near castles can not be occupied.
     * Randomly selects an obstacle type for flavor reasons.
     */
    private void generateFields() {

        for (int i = 0; i < width; i++) {
            for (int j = 0; j < height; j++) {
                setField(i, j, FieldType.OPEN);
            }
        }

        int numberOfObstacles = (int) ((width * height) * 0.2);

        for (int i = 0; i < numberOfObstacles; i++) {
            int x = random.nextInt(width);
            int y = random.nextInt(height);

            while (nearCastle(x, y) || fields.get(x).get(y).getType() != FieldType.OPEN) {
                x = random.nextInt(width);
                y = random.nextInt(height);
            }

            int obstacle = random.nextInt(FieldType.values().length - 1) + 1;
            FieldType type = FieldType.values()[obstacle];

            setField(x, y, type);
        }
    }

    /**
     * Checks if the selected field is in the 3x3 vicinity of a castle.
     *
     * @param x The selected field's column
     * @param y THe selected field's row
     * @return If the field is in the 3x3 vicinity of a castle
     */
    public boolean nearCastle(int x, int y) {
        Point2D position = new Point2D(x, y);
        return coordinatesNearCastles.contains(position);
    }

    /**
     * Checks if the selected field is in the 3x3 vicinity of a castle.
     *
     * @param position The selected field's position
     * @return If the field is in the 3x3 vicinity of a castle
     */
    public boolean nearCastle(Point2D position) {
        return coordinatesNearCastles.contains(position);
    }

    /**
     * Checks if the selected field contains a castle or not.
     *
     * @param x The selected field's column
     * @param y THe selected field's row
     * @return If the field contains a castle or not
     */
    public boolean isCastlePosition(int x, int y) {
        Point2D position = new Point2D(x, y);
        return position.equals(playerACastlePosition) || position.equals(playerBCastlePosition);
    }

    /**
     * Initializes the coordinates near the players' castles.
     */
    private void generateCoordinatesNearCastles() {

        coordinatesNearCastles = new ArrayList<>();
        coordinatesNearCastles.add(playerACastlePosition);
        coordinatesNearCastles.add(playerBCastlePosition);

        List<Point2D> ACastleDirections = List.of(
                new Point2D(0, -1),
                new Point2D(0, 1),
                new Point2D(1, 0),
                new Point2D(1, -1),
                new Point2D(1, 1));

        List<Point2D> BCastleDirections = List.of(
                new Point2D(-1, -1),
                new Point2D(-1, 0),
                new Point2D(-1, 1),
                new Point2D(0, -1),
                new Point2D(0, 1));

        for (Point2D direction : ACastleDirections) {
            Point2D neighbor = new Point2D(playerACastlePosition.getX()+direction.getX(), playerACastlePosition.getY()+direction.getY());
            coordinatesNearCastles.add(neighbor);
        }

        for (Point2D direction : BCastleDirections) {
            Point2D neighbor = new Point2D(playerBCastlePosition.getX()+direction.getX(), playerBCastlePosition.getY()+direction.getY());
            coordinatesNearCastles.add(neighbor);
        }

    }

    /**
     * Adds a tower to the selected field.
     *
     * @param column The desired field's column
     * @param row The desired field's row
     * @param newTower The desired tower to be placed
     */
    public void addTower(int column, int row, Tower newTower) {
        getField(row, column).addTower(newTower);
        Point2D obstaclePosition = new Point2D(column, row);
        pathFinder.removeFromNodes(obstaclePosition);
    }

    /**
     * Adds a building(castle, barrack) to the selected field.
     *
     * @param column The desired field's column
     * @param row The desired field's row
     * @param newBuilding The desired building to be placed
     */
    public void addBuilding(int column, int row, Building newBuilding) {
        getField(row, column).addBuilding(newBuilding);
        Point2D obstaclePosition = new Point2D(column, row);
        pathFinder.removeFromNodes(obstaclePosition);
    }

    /**
     * Returns the positions of fields of the map's first half.
     *
     * @return The list of positions
     */
    public List<Point2D> getPlayerAFieldsPositions() {
        List<Point2D> fieldPositions = new ArrayList<>();
            for (List<Field> row : fields) {
                for (Field field : row) {
                    if ((field.getPosition().getX() < (double)width/2)) {
                        fieldPositions.add(field.getPosition());
                    }
                }
            }
        return fieldPositions;
    }

    /**
     * Returns the positions of fields of the map's second half.
     *
     * @return The list of positions
     */
    public List<Point2D> getPlayerBFieldsPositions() {
        List<Point2D> fieldPositions = new ArrayList<>();
        for (List<Field> row : fields) {
            for (Field field : row) {
                if ((width % 2 == 0 && field.getPosition().getX() >= (double)width/2)
                        || (width % 2 != 0 && field.getPosition().getX() > (double)width/2)) {
                    fieldPositions.add(field.getPosition());
                }
            }
        }
        return fieldPositions;
    }

    /**
     * Returns the positions of the map which are
     * close enough to playerA's castle.
     *
     * @return The list of valid positions
     */
    public List<Point2D> getPlayerAPossibleBarrackPositions() {

        List<Point2D> possibleBarrackPositions = new ArrayList<>();

        for (Point2D p : getPlayerAFieldsPositions()) {
            if (p.getX() <= (double)width/4)
                possibleBarrackPositions.add(p);
        }

        return possibleBarrackPositions;
    }

    /**
     * Returns the positions of the map which are
     * close enough to playerB's castle.
     *
     * @return The list of valid positions
     */
    public List<Point2D> getPlayerBPossibleBarrackPositions() {

        List<Point2D> possibleBarrackPositions = new ArrayList<>();

        for (Point2D p : getPlayerBFieldsPositions()) {
            if (p.getX() >= width - (double)width/4 - 1)
                possibleBarrackPositions.add(p);
        }

        return possibleBarrackPositions;
    }

    @Override
    public String toString() {
        return getString(fields);
    }

    public static String getString(List<List<Field>> fields) {
        StringBuilder line = new StringBuilder();
        for (List<Field> row : fields) {
            for (Field field : row) {
                line.append(field.getType());
                line.append(" ");
            }
            line.append("\n");
        }
        return line.toString();
    }
}
