package bitsplease.mummymayhem.model.units;

import bitsplease.mummymayhem.model.buildings.Building;
import bitsplease.mummymayhem.model.buildings.Tower;
import bitsplease.mummymayhem.model.map.Field;
import bitsplease.mummymayhem.model.map.Map;
import bitsplease.mummymayhem.model.utils.Direction;
import bitsplease.mummymayhem.model.utils.PathFinder;
import bitsplease.mummymayhem.model.utils.Sprite;
import javafx.geometry.Point2D;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;

public abstract class Unit extends Sprite {

    protected final int initialHp;

    protected int hp;
    protected int speed;
    protected int attack;
    protected int price;
    protected List<Field> path;
    protected List<Point2D> checkPoints;

    protected String imageFolderPath;

    protected double distanceFromDestination;
    protected Field destination;
    protected Direction facing;
    protected int currentAnimationPhase;
    protected int nextAnimationPhase;
    protected List<List<String>> animationStages;
    protected long lastMovementUpdate = 0;
    protected long lastAnimationUpdate = 0;
    protected int priority;

    protected Unit(Point2D position, String imageFolderPath, int hp, int speed, int attack, int price) {
        super(position, "");
        this.hp = hp;
        this.speed = speed;
        this.attack = attack;
        this.price = price;
        path = new ArrayList<>();
        checkPoints = new ArrayList<>();

        this.imageFolderPath = imageFolderPath;

        this.initialHp = this.hp;

        this.animationStages = new ArrayList<>(4);
        for(int i = 0; i < 4; i++) {
            this.animationStages.add(new ArrayList<>(3));
        }
    }

    protected Unit(Unit other) {
        super(other.position, other.imagePath);
        this.hp = other.hp;
        this.speed = other.speed;
        this.attack = other.attack;
        this.price = other.price;

        this.imageFolderPath = other.imageFolderPath;

        this.destination = other.destination;
        this.distanceFromDestination = other.distanceFromDestination;
        this.facing = other.facing;
        this.currentAnimationPhase = other.currentAnimationPhase;
        this.nextAnimationPhase = other.nextAnimationPhase;
        this.animationStages = other.animationStages;
        this.lastMovementUpdate = other.lastMovementUpdate;
        this.lastAnimationUpdate = other.lastAnimationUpdate;
        this.priority = other.priority;

        this.initialHp = this.hp;

        path = new ArrayList<>();
        checkPoints = new ArrayList<>();
    }

    /**
     * Decreases the unit's hp by a given value,
     * or sets hp to 0 if the value is higher than current hp.
     *
     * @param damage The value of the incoming damage
     */
    public void takeDamage(int damage) {
        if(hp <= damage) {
            hp = 0;
        }
        else {
            hp -= damage;
        }
    }

    public abstract String getName();
    /**
     * Attacks the target building(Barrack, Castle), decreasing its health by the unit's attack value
     *
     * @param building The target building
     */

    public void attack(Building building) {
        building.takeDamage(this.attack);
    }

    public void attack(Tower tower) {
        tower.takeDamage(this.attack);
    }

    /**
     * Moves the unit on the X in the given direction
     *
     * @param direction The direction of the move
     */
    public void moveX(double direction) {
        this.position = position.add(direction, 0);
        this.distanceFromDestination -= Math.abs(direction);
    }


    /**
     * Moves the unit on the Y axis in the given direction
     *
     * @param direction The direction of the move
     */
    public void moveY(double direction) {
        this.position = position.add(0, direction);
        this.distanceFromDestination -= Math.abs(direction);
    }

    /**
     * Adds a field to the unit's path
     *
     * @param field The desired field to add
     */
    public void addFieldToPath(Field field) {
        path.add(field);
    }

    /**
     * Initializes the image arrays with the right images.
     */
    public void initAnimationStages() {
        for(int i = 0; i < 4; i++) {
            for(int j = 1; j <= 3; j++) {
                StringBuilder direction = new StringBuilder();
                switch (i) {
                    case 0 -> direction.append("up");
                    case 1 -> direction.append("right");
                    case 2 -> direction.append("down");
                    case 3 -> direction.append("left");
                }
                String directionString = direction.toString();
                this.animationStages.get(i).add(this.imageFolderPath + directionString + "_" + j + ".png");
            }
        }
    }

    /**
     * Calculates and sets the default path without checkpoints,
     * only start position and destination are given.
     *
     * @param map The map object
     * @param destination The destination's position
     */
    public void calculateDefaultPath(Map map, Point2D destination) {

        PathFinder pathFinder = new PathFinder(map.getWidth(), map.getHeight(), position, destination, map.getFields());
        List<Field> newPath = new ArrayList<>();

        for (Point2D p : pathFinder.getPath()) {
            newPath.add(map.getFields().get((int)p.getX()).get((int)p.getY()));
        }
        this.setPath(newPath);
    }

    /**
     * Can be called if the unit already has a default path.
     * Generate a new path using start position, destination
     * and checkpoints, then returns it in a list of positions.
     *
     * @param map The map object
     * @return The new path represented by a list of points
     */
    public List<Point2D> getUpdatedPath(Map map) {

        List<Point2D> newPath = new ArrayList<>();
        Point2D from = this.position;
        Point2D destination = this.destination.getPosition();
        PathFinder pathFinder;

        for (Point2D checkPoint : this.checkPoints) {
            pathFinder = new PathFinder(map.getWidth(), map.getHeight(), from, checkPoint, map.getFields());
            newPath.addAll(pathFinder.getPath());
            newPath.remove(newPath.size()-1);
            from = checkPoint;
        }

        pathFinder = new PathFinder(map.getWidth(), map.getHeight(), from, destination, map.getFields());
        newPath.addAll(pathFinder.getPath());

        return newPath;

    }

    /**
     * Deletes checkpoints which became unreachable.
     *
     * @param map The map object
     */
    public void updateCheckpoints(Map map) {
        checkPoints.removeIf(checkpoint -> !map.getPathFinder().isTraversable(checkpoint));
    }

    public int getHp() {
        return hp;
    }

    public void setHp(int hp) {
        this.hp = hp;
    }

    public int getInitialHp() {
        return initialHp;
    }

    public int getSpeed() {
        return speed;
    }

    public int getAttack() {
        return attack;
    }

    public void setAttack(int attack) {
        this.attack = attack;
    }

    public int getPrice() {
        return price;
    }

    public void setPrice(int price) {
        this.price = price;
    }

    public List<Field> getPath() {
        return path;
    }

    public void setPath(List<Field> newPath) {
        this.path = new ArrayList<>(newPath);
        this.distanceFromDestination = this.path.size() - 1;
        this.destination = path.get(path.size()-1);
    }

    public String getImageFolderPath() {
        return imageFolderPath;
    }

    public int getCurrentAnimationPhase() {
        return currentAnimationPhase;
    }

    public int getNextAnimationPhase() {
        return nextAnimationPhase;
    }

    public Field getDestination() {
        return destination;
    }

    public double getDistanceFromDestination() {
        return distanceFromDestination;
    }

    public List<String> getUpwardsFacingImages() {
        return animationStages.get(0);
    }

    public List<String> getRightFacingImages() {
        return animationStages.get(1);
    }

    public List<String> getDownwardsFacingImages() {
        return animationStages.get(2);
    }

    public List<String> getLeftFacingImages() {
        return animationStages.get(3);
    }

    public long getLastMovementUpdate() {
        return lastMovementUpdate;
    }

    public void setLastMovementUpdate(long lastMovementUpdate) {
        this.lastMovementUpdate = lastMovementUpdate;
    }

    public long getLastAnimationUpdate() {
        return lastAnimationUpdate;
    }

    public void setLastAnimationUpdate(long lastAnimationUpdate) {
        this.lastAnimationUpdate = lastAnimationUpdate;
    }

    public int getPriority() {
        return priority;
    }

    public void setPriority(int priority) {
        this.priority = priority;
    }

    public Direction getFacing() {
        return facing;
    }

    /**
     * Turns the unit in a given direction.
     *
     * @param facing The desired direction to turn the unit.
     */
    public void setFacing(Direction facing) {
        if(this.facing == null || !this.facing.equals(facing)) {
            this.facing = facing;
            currentAnimationPhase = 0;
            nextAnimationPhase = 1;
            nextAnimationImage();
        }
    }

    /**
     * Sets the unit's image to the one that matches the direction and animation phase.
     */
    public void nextAnimationImage() {
        switch (facing) {
            case UP -> this.imagePath = getUpwardsFacingImages().get(currentAnimationPhase);
            case RIGHT -> this.imagePath = getRightFacingImages().get(currentAnimationPhase);
            case DOWN -> this.imagePath = getDownwardsFacingImages().get(currentAnimationPhase);
            case LEFT -> this.imagePath = getLeftFacingImages().get(currentAnimationPhase);
        }
    }

    /**
     * Updates the unit's animation phase,
     * and decides what phase to set after.
     */
    public void nextAnimationPhase() {
        int oldAnimationPhase = currentAnimationPhase;
        currentAnimationPhase = nextAnimationPhase;

        switch(oldAnimationPhase) {
            case 0 -> nextAnimationPhase = 2;
            case 1 -> nextAnimationPhase = 1;
            case 2 -> nextAnimationPhase = 0;
        }

        nextAnimationImage();
    }

    /**
     * Checks what direction the upcoming field in the unit's path is.
     *
     * @return The direction the upcoming field un the unit's path is.
     */
    public Direction nextFieldDirection() {
        position = new Point2D(Math.round(position.getX()), Math.round(position.getY()));
        distanceFromDestination = Math.round(distanceFromDestination);

        Point2D nextFieldPos = path.get(0).getPosition();
        if(position.getX() == nextFieldPos.getX() && position.getY() > nextFieldPos.getY()) {
            return Direction.UP;
        } else if (position.getX() == nextFieldPos.getX() && position.getY() < nextFieldPos.getY()) {
            return Direction.DOWN;
        } else if (position.getX() > nextFieldPos.getX() && position.getY() == nextFieldPos.getY()) {
            return Direction.LEFT;
        } else {
            return Direction.RIGHT;
        }
    }

    /**
     * Checks if the unit reached the next field in its path,
     * if so, it turns towards the field after.
     */
    public void checkNextField() {
        if(facing == null) {
            setFacing(nextFieldDirection());
            return;
        }
        Point2D nextFieldPos = path.get(0).getPosition();

        if ((facing.equals(Direction.UP) && position.getY() <= nextFieldPos.getY())
                || (facing.equals(Direction.DOWN) && position.getY() >= nextFieldPos.getY())
                || (facing.equals(Direction.RIGHT) && position.getX() >= nextFieldPos.getX())
                || (facing.equals(Direction.LEFT) && position.getX() <= nextFieldPos.getX())) {
            removeNextField();
            if(path.size() >= 1) {
                setFacing(nextFieldDirection());
            }
        }
    }

    /**
     * Checks if the unit reached the other player's castle.
     *
     * @return true if the unit reached the other player's castle, false if not.
     */
    public boolean isDestinationReached() {
        return (facing.equals(Direction.UP) && position.getY() <= destination.getPosition().getY())
                || (facing.equals(Direction.DOWN) && position.getY() >= destination.getPosition().getY())
                || (facing.equals(Direction.RIGHT) && position.getX() >= destination.getPosition().getX())
                || (facing.equals(Direction.LEFT) && position.getX() <= destination.getPosition().getX());
    }

    /**
     * Removes the first field from the unit's path.
     */
    public void removeNextField() {
        path.remove(0);
    }

    /**
     * Adds a checkpoint to the checkpoint's collection,
     * then sort it by their distance from unit's start position.
     *
     * @param position The checkpoint
     */
    public void addCheckPoint(Point2D position) {
        this.checkPoints.add(position);
        Point2D start = this.position;
        checkPoints.sort(Comparator.comparing((Point2D p) -> p.distance(start)));
    }

    /**
     * Deletes a checkpoint from the checkpoint's collection.
     *
     * @param position The checkpoint
     */
    public void deleteCheckPoint(Point2D position) {
        this.checkPoints.remove(position);
    }

    public List<Point2D> getCheckPoints() {
        return checkPoints;
    }

    public void setCheckPoints(List<Point2D> positions) {
        this.checkPoints = new ArrayList<>(positions);
    }
}
