package bitsplease.mummymayhem.model.utils;

import bitsplease.mummymayhem.model.map.Field;
import javafx.geometry.Point2D;
import java.util.*;
import java.util.List;

/*
 * https://www.redblobgames.com/pathfinding/tower-defense/
 */

public class PathFinder {

    protected final int WIDTH;
    protected final int HEIGHT;
    protected final Point2D START;
    protected final Point2D DESTINATION;

    protected List<Point2D> allNodes = new ArrayList<>();
    protected Map<Point2D, Point2D> cameFrom = new HashMap<>();

    public PathFinder(int width, int height, Point2D start, Point2D destination, List<List<Field>> fields) {
        this.WIDTH = width;
        this.HEIGHT = height;
        this.START = start;
        this.DESTINATION = destination;
        refresh(fields);
    }

    /**
     * Adds open fields to 'allNodes'.
     *
     * @param fields The map's fields.
     */
    protected void setFreeNodes(List<List<Field>> fields) {

        allNodes = new ArrayList<>();
        for (int i = 0; i < WIDTH; i++) {
            for (int j = 0; j < HEIGHT; j++) {
                if (fields.get(i).get(j).getPosition().equals(DESTINATION) || !fields.get(i).get(j).isOccupied()) {
                    allNodes.add(new Point2D(i, j));
                }
            }
        }
    }

    /**
     * Adds reached fields from start position to 'cameFrom'.
     */
    protected void generateReachedNodesFromStart() {

        Queue<Point2D> frontier = new LinkedList<>();
        cameFrom = new HashMap<>();

        frontier.add(START);
        cameFrom.put(START, null);

        while (frontier.size() != 0) {
            Point2D current = frontier.poll();
            for (Point2D next : neighbors(current)) {
                if (!cameFrom.containsKey(next)) {
                    frontier.add(next);
                    cameFrom.put(next, current);
                }
            }
        }
    }

    /**
     * Checks whether we can walk from start position to the given destination or not.
     *
     * @param destination The position to reach
     * @return true if map is traversable, false otherwise
     */
    public boolean isTraversable(Point2D destination) {
        return cameFrom.containsKey(destination);
    }

    /**
     * Checks whether we can walk from start position to destination or not.
     *
     * @return true if map is traversable, false otherwise
     */
    public boolean isTraversable() {
        return cameFrom.containsKey(DESTINATION);
    }

    /**
     * Searches for the given position's neighbors,
     * and collects open fields from them to 'result'.
     *
     * @param node The position
     * @return The list of the position's reachable neighbors
     */
    protected List<Point2D> neighbors(Point2D node) {

        List<Point2D> result = new ArrayList<>();
        List<Point2D> directions = List.of(
                new Point2D(1, 0),
                new Point2D(0, 1),
                new Point2D(-1, 0),
                new Point2D(0, -1));

        for (Point2D direction : directions) {
            Point2D neighbor = new Point2D(node.getX() + direction.getX(), node.getY() + direction.getY());
            if (allNodes.contains(neighbor)) {
                result.add(neighbor);
            }
        }

        return result;
    }

    /**
     * Removes the given position from the open nodes,
     * then regenerates reached nodes and checks if map
     * is still traversable.
     * Finally, reverses changes.
     *
     * @param obstaclePosition The position to delete from open nodes
     * @return true if map remains traversable, false otherwise
     */
    public boolean blockPath(Point2D obstaclePosition) {

        removeFromNodes(obstaclePosition);

        //check if map is still traversable
        boolean result = isTraversable();

        //reverse changes
        allNodes.add(obstaclePosition);
        generateReachedNodesFromStart();

        return !result;
    }

    /**
     * Sets open fields and generates reached fields from start position.
     *
     * @param fields The map's fields
     */
    public void refresh(List<List<Field>> fields) {
        setFreeNodes(fields);
        generateReachedNodesFromStart();
    }

    /**
     * Removes the given position from the open nodes,
     * then regenerates reached nodes.
     *
     * @param obstaclePosition The position to delete
     */
    public void removeFromNodes(Point2D obstaclePosition) {
        //remove obstaclePosition from the free path collection
        allNodes.remove(obstaclePosition);
        //regenerate reached nodes
        generateReachedNodesFromStart();
    }


    /**
     * Returns the shortest path between START and DESTINATION.
     *
     * @return The list of the shortest path's positions.
     */
    public List<Point2D> getPath() {

        List<Point2D> path = new ArrayList<>();
        Point2D position = DESTINATION;

        while (!position.equals(START)) {
            path.add(position);
            position = cameFrom.get(position);
        }
        path.add(START);
        Collections.reverse(path);

        return path;
    }

}
