package bitsplease.mummymayhem.model.utils;

import bitsplease.mummymayhem.model.GameModel;
import bitsplease.mummymayhem.model.PhaseType;
import bitsplease.mummymayhem.model.PlayerType;
import bitsplease.mummymayhem.model.buildings.Barrack;
import bitsplease.mummymayhem.model.buildings.Tower;
import bitsplease.mummymayhem.model.map.Field;
import bitsplease.mummymayhem.model.map.Map;
import bitsplease.mummymayhem.model.units.Unit;
import javafx.geometry.Point2D;
import javafx.geometry.Rectangle2D;
import javafx.scene.canvas.Canvas;
import javafx.scene.canvas.GraphicsContext;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.paint.Color;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;

public class Renderer {

    private Map map;
    private final GraphicsContext graphics;
    private double widthRatio, heightRatio;

    public Renderer(Map map, GraphicsContext graphics) {
        this.graphics = graphics;
        this.map = map;
    }

    public void setWidthRatio(double widthRatio) {
        this.widthRatio = widthRatio;
    }

    public void setHeightRatio(double heightRatio) {
        this.heightRatio = heightRatio;
    }

    public void setMap(Map map) {
        this.map = map;
    }

    /**
     * Renders the fields one by one according to their positions.
     */
    public void drawFields() {
        for(List<Field> i : map.getFields()) {
            for(Field j : i) {
                graphics.drawImage(ImageStorage.getImage(j), j.getPosition().getX()* widthRatio, j.getPosition().getY()* heightRatio, widthRatio, heightRatio);
            }
        }
    }

    /**
     * Renders both players' castles on the correct position.
     */
    public void drawCastles() {

        double playerACastleX = map.getPlayerACastlePosition().getX();
        double playerACastleY = map.getPlayerACastlePosition().getY();
        double playerBCastleX = map.getPlayerBCastlePosition().getX();
        double playerBCastleY = map.getPlayerBCastlePosition().getY();

        graphics.drawImage(ImageStorage.getImage(PlayerType.PLAYER_A.getImagePath()), playerACastleX* widthRatio, playerACastleY* heightRatio, widthRatio, heightRatio);
        graphics.drawImage(ImageStorage.getImage(PlayerType.PLAYER_B.getImagePath()), playerBCastleX* widthRatio, playerBCastleY* heightRatio, widthRatio, heightRatio);
    }

    /**
     * Renders both players' towers, and their projectiles on the correct positions.
     */
    public void drawTowers(GameModel model) {
        List<Tower> playerATowers = model.getPlayerA().getTowers();
        List<Tower> playerBTowers = model.getPlayerB().getTowers();

        List<Tower> towers = new ArrayList<>();

        towers.addAll(playerATowers);
        towers.addAll(playerBTowers);

        for(Tower t: towers) {
            double towerX = t.getPosition().getX();
            double towerY = t.getPosition().getY();

            graphics.drawImage(ImageStorage.getImage(t), towerX * widthRatio, towerY * heightRatio, widthRatio, heightRatio);

            double projectileX = t.getProjectilePosition().getX();
            double projectileY = t.getProjectilePosition().getY();

            if(projectileX != towerX || projectileY != towerY) {
                graphics.drawImage(ImageStorage.getImage(t.getProjectileFilePath()), projectileX * widthRatio + (widthRatio/2 - widthRatio/10), projectileY * heightRatio + (heightRatio/2 - heightRatio/10), widthRatio/10, heightRatio/10);
            }
        }
    }

    /**
     * Renders both players' barracks on the correct positions.
     */
    public void drawBarracks(GameModel model) {
        List<Barrack> playerABarracks = model.getPlayerA().getBarracks();
        List<Barrack> playerBBarracks = model.getPlayerB().getBarracks();

        List<Barrack> barracks = new ArrayList<>();

        barracks.addAll(playerABarracks);
        barracks.addAll(playerBBarracks);

        for(Barrack b: barracks) {
            double barrackX = b.getPosition().getX();
            double barrackY = b.getPosition().getY();

            graphics.drawImage(ImageStorage.getImage(b), barrackX * widthRatio, barrackY * heightRatio, widthRatio, heightRatio);
        }
    }

    public void drawUnits(GameModel model) {
        if(!model.getPhase().equals(PhaseType.ATTACK)) { return; }

        List<Unit> playerAUnits = model.getPlayerA().getUnits();
        List<Unit> playerBUnits = model.getPlayerB().getUnits();

        List<Unit> units = new ArrayList<>();

        units.addAll(playerAUnits);
        units.addAll(playerBUnits);

        for(Unit u: units) {

            double unitX = u.getPosition().getX();
            double unitY = u.getPosition().getY();

            Image image = ImageStorage.getImage(u.getImagePath());

            graphics.drawImage(image, unitX * widthRatio, unitY * heightRatio, widthRatio, heightRatio);
        }
    }

    /**
     * Renders the grid of the fields.
     * Draws vertical lines with a spacing equal to a field's width between them,
     * draws horizontal lines with a spacing equal to a field's height between them.
     */
    public void drawGrid(Canvas gameCanvas) {
        graphics.setStroke(Color.DARKORANGE);
        for(double i=0d; i <= gameCanvas.getHeight(); i += heightRatio) {
            graphics.beginPath();
            graphics.moveTo(0d, i);
            graphics.lineTo(gameCanvas.getWidth(), i);
            graphics.stroke();
        }
        for(double i=0d; i <= gameCanvas.getWidth(); i += widthRatio) {
            graphics.beginPath();
            graphics.moveTo(i, 0d);
            graphics.lineTo(i, gameCanvas.getHeight());
            graphics.stroke();
        }
    }

    /**
     * Draws the line of the selected unit's path,
     * and draws checkpoints as filled circles.
     */
    public void drawUnitPath(GameModel model, Unit selectedUnit, Point2D selectedUnitPosition) {

        if (selectedUnit != null) {
            graphics.setStroke(Color.BURLYWOOD);
            graphics.setFill(Color.BURLYWOOD);

            //search the unit (selectedUnit is just a sample)
            Optional<Unit> unit = model.getActualPlayer().getUnits().stream()
                    .filter(u -> u.getClass().getName().equals(selectedUnit.getClass().getName()))
                    .filter(u -> u.getPosition().equals(selectedUnitPosition))
                    .findFirst();

            if (unit.isPresent()) {
                graphics.beginPath();

                //draw path
                graphics.moveTo(unit.get().getPosition().getX()*widthRatio + widthRatio/2,
                        unit.get().getPosition().getY()*heightRatio + heightRatio/2);
                graphics.setLineWidth(3);
                graphics.setLineDashes(5);

                for (Field field : unit.get().getPath()) {
                    double x = field.getPosition().getX();
                    double y = field.getPosition().getY();
                    graphics.lineTo(x*widthRatio + widthRatio/2, y*heightRatio + heightRatio/2);
                    graphics.stroke();
                }

                //draw checkpoints
                graphics.setLineDashes(0);
                for (Point2D checkpoint : unit.get().getCheckPoints()) {
                    graphics.strokeOval(checkpoint.getX() * widthRatio + widthRatio / 2 - 5,
                            checkpoint.getY() * heightRatio + heightRatio/2 - 5, 10, 10);

                    graphics.fillOval(checkpoint.getX() * widthRatio + widthRatio / 2 - 5,
                            checkpoint.getY() * heightRatio + heightRatio/2 - 5, 10, 10);
                }
            }
        }
    }

    /**
     * Draws castles', towers' and units' (in case of attack phase) hp bar.
     *
     * @param model The GameModel object
     */
    public void drawHpBars(GameModel model) {

        drawCastlesHpBar(model);
        drawTowersHpBar(model);
        if(model.getPhase().equals(PhaseType.ATTACK)) {
            drawUnitsHpBar(model);
        }
    }

    /**
     * Paints the canvas to black with 0.8 opacity.
     */
    public void darkenCanvas() {
        graphics.setFill(Color.BLACK);
        graphics.setGlobalAlpha(0.8);
        graphics.fillRect(0, 0, map.getWidth()*widthRatio, map.getHeight()*heightRatio);
        graphics.setGlobalAlpha(1.0);
    }

    /**
     * Draws a hp bar with the given values.
     *
     * @param path The appropriate image's path
     * @param posX The position x
     * @param posY The position y
     * @param remainingHpProportion The remained hp proportion
     * @param width The hp bar's width proportion (according to widthRatio)
     * @param height The hp bar's height proportion (according to heightRatio)
     * @param heightShift The shift proportion (according to heightRatio)
     */
    private void drawHpBar(String path, double posX, double posY, double remainingHpProportion, double width, double height, double heightShift) {
        Image background = ImageStorage.getImage("src/main/resources/assets/images/health-bar-bg.png");
        Image fullImage = ImageStorage.getImage(path);

        graphics.drawImage(background,
                posX * widthRatio + (widthRatio*(1-width))/2,
                posY * heightRatio - heightRatio*heightShift,
                widthRatio*width, heightRatio*height);

        Rectangle2D croppedPortion = new Rectangle2D(
                fullImage.getWidth()*remainingHpProportion,
                0,
                fullImage.getWidth(),
                fullImage.getHeight());

        ImageView imageView = new ImageView(fullImage);
        imageView.setViewport(croppedPortion);
        imageView.setSmooth(true);

        graphics.drawImage(imageView.getImage(),
                posX * widthRatio + (widthRatio*(1-width))/2,
                posY * heightRatio - heightRatio*heightShift,
                widthRatio*width*remainingHpProportion, heightRatio*height);
    }

    /**
     * Draws hp bars for castles.
     *
     * @param model The GameModel object
     */
    private void drawCastlesHpBar(GameModel model) {

        //Castle A
        double remainingHpProportionA = (double)model.getPlayerA().getCastle().getHp() / model.getPlayerA().getCastle().getInitialHp();

        drawHpBar("src/main/resources/assets/images/health-bar-playerA.png",
                model.getPlayerA().getCastle().getPosition().getX(),
                model.getPlayerA().getCastle().getPosition().getY(),
                remainingHpProportionA, 0.9,  0.1, 0.15);

        //Castle B
        double remainingHpProportionB = (double)model.getPlayerB().getCastle().getHp() / model.getPlayerB().getCastle().getInitialHp();

        drawHpBar("src/main/resources/assets/images/health-bar-playerB.png",
                model.getPlayerB().getCastle().getPosition().getX(),
                model.getPlayerB().getCastle().getPosition().getY(),
                remainingHpProportionB, 0.9,  0.1, 0.15);
    }

    /**
     * Draws hp bars for towers.
     *
     * @param model The GameModel object
     */
    private void drawTowersHpBar(GameModel model) {

        //Player A
        for (Tower tower : model.getPlayerA().getTowers()) {

            double remainingHpProportion = (double) tower.getHp() / tower.getInitialHp();

            drawHpBar("src/main/resources/assets/images/health-bar-playerA.png",
                    tower.getPosition().getX(),
                    tower.getPosition().getY(),
                    remainingHpProportion, 0.8,  0.1, 0.15);
        }

        //Player B
        for (Tower tower : model.getPlayerB().getTowers()) {
            double remainingHpProportion = (double) tower.getHp() / tower.getInitialHp();

            drawHpBar("src/main/resources/assets/images/health-bar-playerB.png",
                    tower.getPosition().getX(),
                    tower.getPosition().getY(),
                    remainingHpProportion, 0.8,  0.1, 0.15);
        }
    }

    /**
     * Draws hp bars for units if they have less hp
     * than initial hp.
     *
     * @param model The GameModel object
     */
    private void drawUnitsHpBar(GameModel model) {

        //Player A
        for (Unit unit : model.getPlayerA().getUnits()) {

            double remainingHpProportion = (double)unit.getHp() / unit.getInitialHp();

            if (remainingHpProportion < 1) {

                drawHpBar("src/main/resources/assets/images/health-bar-playerA.png",
                        unit.getPosition().getX(),
                        unit.getPosition().getY(),
                        remainingHpProportion, 0.4,  0.08, 0.05);
            }
        }

        //Player B
        for (Unit unit : model.getPlayerB().getUnits()) {

            double remainingHpProportion = (double)unit.getHp() / unit.getInitialHp();

            if (remainingHpProportion < 1) {

                drawHpBar("src/main/resources/assets/images/health-bar-playerB.png",
                        unit.getPosition().getX(),
                        unit.getPosition().getY(),
                        remainingHpProportion, 0.4,  0.08, 0.05);
            }
        }
    }
}
