package bitsplease.mummymayhem.model;

import bitsplease.mummymayhem.model.buildings.Barrack;
import bitsplease.mummymayhem.model.buildings.Tower;
import bitsplease.mummymayhem.model.buildings.TowerType;
import bitsplease.mummymayhem.model.map.Map;
import bitsplease.mummymayhem.model.units.Demolisher;
import bitsplease.mummymayhem.model.units.Runner;
import bitsplease.mummymayhem.model.units.Unit;
import bitsplease.mummymayhem.model.utils.Sprite;
import javafx.geometry.Point2D;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.util.Collection;
import java.util.List;
import java.util.Optional;

public class GameModelTest {

    GameModel gameModel;

    @BeforeEach
    void setUp() {
        gameModel = new GameModel();
    }

    private Point2D getRandomOpenFieldFromOwnHalf(PlayerType type) {
        Map map = gameModel.getMap();
        List<Point2D> ownHalf = type == PlayerType.PLAYER_A ?
                gameModel.getMap().getPlayerAFieldsPositions() :
                gameModel.getMap().getPlayerBFieldsPositions();

        Optional<Point2D> position = map.getFields().stream()
                .flatMap(Collection::stream)
                .filter(f -> !f.isOccupied())
                .map(Sprite::getPosition)
                .filter(ownHalf::contains)
                .findAny();

        return position.orElse(null);
    }

    private Point2D getRandomOpenFieldFromOwnQuarter(PlayerType type) {
        Map map = gameModel.getMap();
        List<Point2D> ownQuarter = type == PlayerType.PLAYER_A ?
                gameModel.getMap().getPlayerAPossibleBarrackPositions() :
                gameModel.getMap().getPlayerBPossibleBarrackPositions();

        Optional<Point2D> position = map.getFields().stream()
                .flatMap(Collection::stream)
                .filter(f -> !f.isOccupied())
                .map(Sprite::getPosition)
                .filter(ownQuarter::contains)
                .findAny();

        return position.orElse(null);
    }

    @Test
    void nextPhaseTest() {

        PhaseType.reset();
        Assertions.assertEquals(PhaseType.PLAYER_A_BUILD, gameModel.getPhase());

        gameModel.nextPhase();
        Assertions.assertEquals(PhaseType.PLAYER_B_BUILD, gameModel.getPhase());

        gameModel.nextPhase();
        Assertions.assertEquals(PhaseType.ATTACK, gameModel.getPhase());

        gameModel.nextPhase();
        Assertions.assertEquals(PhaseType.PLAYER_B_BUILD, gameModel.getPhase());

        gameModel.nextPhase();
        Assertions.assertEquals(PhaseType.PLAYER_A_BUILD, gameModel.getPhase());

        gameModel.nextPhase();
        Assertions.assertEquals(PhaseType.ATTACK, gameModel.getPhase());
    }

    @Test
    void getActualPlayerTest() {

        PhaseType.reset();
        Assertions.assertEquals(gameModel.getPlayerA(), gameModel.getActualPlayer());

        gameModel.nextPhase();
        Assertions.assertEquals(gameModel.getPlayerB(), gameModel.getActualPlayer());

        gameModel.nextPhase();
        gameModel.nextPhase();
        Assertions.assertEquals(gameModel.getPlayerB(), gameModel.getActualPlayer());

        gameModel.nextPhase();
        Assertions.assertEquals(gameModel.getPlayerA(), gameModel.getActualPlayer());
    }

    @Test
    void selectBuildingTest() {
        Tower tower = new Tower(TowerType.BLOCKADE);

        gameModel.selectBuilding(tower);
        Assertions.assertEquals(tower, gameModel.getSelectedBuilding());

        gameModel.selectBuilding(null);
        Assertions.assertNull(gameModel.getSelectedBuilding());
    }

    @Test
    void getWinnerTest() {
        gameModel.getPlayerA().getCastle().setHp(0);
        Assertions.assertEquals(PlayerType.PLAYER_B, gameModel.getWinner());
    }

    @Test
    void getWinnerTest2() {
        gameModel.getPlayerB().getCastle().setHp(0);
        Assertions.assertEquals(PlayerType.PLAYER_A, gameModel.getWinner());
    }

    @Test
    void getWinnerTest3() {
        gameModel.getPlayerA().getCastle().setHp(0);
        gameModel.getPlayerB().getCastle().setHp(0);
        Assertions.assertNull(gameModel.getWinner());
    }

    @Test
    void gameEndedTest() {
        gameModel.getPlayerA().getCastle().setHp(0);
        Assertions.assertTrue(gameModel.gameEnded());
    }

    @Test
    void gameEndedTest2() {
        gameModel.getPlayerB().getCastle().setHp(0);
        Assertions.assertTrue(gameModel.gameEnded());
    }

    @Test
    void playerReachedBarrackLimitTest() {
        Barrack barrack1 = new Barrack(new Point2D(0,0));
        Barrack barrack2 = new Barrack(new Point2D(0,0));
        Barrack barrack3 = new Barrack(new Point2D(0,0));

        gameModel.getActualPlayer().buyBarrack(barrack1);
        gameModel.getActualPlayer().buyBarrack(barrack2);
        Assertions.assertFalse(gameModel.playerReachedBarrackLimit());

        gameModel.getActualPlayer().buyBarrack(barrack3);
        Assertions.assertTrue(gameModel.playerReachedBarrackLimit());
    }

    @Test
    void isBarrackPositionTest() {
        Barrack barrack = new Barrack(new Point2D(1,2));
        gameModel.getPlayerA().buyBarrack(barrack);

        Assertions.assertTrue(gameModel.isBarrackPosition(1, 2));
    }

    @Test
    void isBarrackPositionTest2() {
        Barrack barrack = new Barrack(new Point2D(1,2));
        gameModel.getPlayerA().buyBarrack(barrack);

        Assertions.assertFalse(gameModel.isBarrackPosition(2, 1));
    }

    @Test
    void isTowerPositionTest() {
        Tower tower = new Tower(new Point2D(1,2), TowerType.BLOCKADE);
        gameModel.getPlayerA().buyTower(tower);

        Assertions.assertTrue(gameModel.isTowerPosition(1, 2));
    }

    @Test
    void isTowerPositionTest2() {
        Tower tower = new Tower(new Point2D(1,2), TowerType.BLOCKADE);
        gameModel.getPlayerA().buyTower(tower);

        Assertions.assertFalse(gameModel.isTowerPosition(2, 1));
    }

    @Test
    void isValidTowerPositionTest() {
        Point2D position = getRandomOpenFieldFromOwnHalf(PlayerType.PLAYER_A);
        Assertions.assertTrue(gameModel.isValidTowerPosition((int)position.getX(), (int)position.getY()));
    }

    @Test
    void isValidTowerPositionTest2() {
        Point2D position = getRandomOpenFieldFromOwnHalf(PlayerType.PLAYER_B);
        Assertions.assertFalse(gameModel.isValidTowerPosition((int)position.getX(), (int)position.getY()));
    }

    @Test
    void isValidBarrackPositionTest() {
        Point2D position = getRandomOpenFieldFromOwnQuarter(PlayerType.PLAYER_A);
        Assertions.assertTrue(gameModel.isValidBarrackPosition((int)position.getX(), (int)position.getY()));
    }

    @Test
    void isValidBarrackPositionTest2() {
        Point2D position = getRandomOpenFieldFromOwnQuarter(PlayerType.PLAYER_B);
        Assertions.assertFalse(gameModel.isValidBarrackPosition((int)position.getX(), (int)position.getY()));
    }

    @Test
    void playerBuyUnitTest() {
        Unit unitSample = new Runner(new Point2D(0,0), "test.png");
        Unit boughtUnit = gameModel.playerBuyUnit(unitSample, gameModel.getActualPlayer().getCastle().getPosition());

        Assertions.assertTrue(gameModel.getActualPlayer().getUnits().contains(boughtUnit));
    }

    @Test
    void modifyUnitPath() {
        Unit unitSample = new Runner(new Point2D(0,0), "test.png");
        Unit boughtUnit = gameModel.playerBuyUnit(unitSample, gameModel.getActualPlayer().getCastle().getPosition());

        Point2D checkPoint = gameModel.getActualPlayer().getCastle().getPosition().add(0,1);
        Point2D checkPoint2 = gameModel.getActualPlayer().getCastle().getPosition().add(0,-1);

        gameModel.modifyUnitPath(unitSample, gameModel.getActualPlayer().getCastle().getPosition(),
                                (int)checkPoint.getX(), (int)checkPoint.getY());
        gameModel.modifyUnitPath(unitSample, gameModel.getActualPlayer().getCastle().getPosition(),
                (int)checkPoint2.getX(), (int)checkPoint2.getY());

        Assertions.assertAll(
                () -> Assertions.assertTrue(boughtUnit.getCheckPoints().contains(checkPoint)),
                () -> Assertions.assertTrue(boughtUnit.getCheckPoints().contains(checkPoint2))
        );
    }

    @Test
    void modifyUnitPath2() {
        Unit unitSample = new Runner(new Point2D(0,0), "test.png");
        Unit boughtUnit = gameModel.playerBuyUnit(unitSample, gameModel.getActualPlayer().getCastle().getPosition());

        Point2D checkPoint = gameModel.getActualPlayer().getCastle().getPosition().add(0,1);

        gameModel.modifyUnitPath(unitSample, gameModel.getActualPlayer().getCastle().getPosition(),
                (int)checkPoint.getX(), (int)checkPoint.getY());
        gameModel.modifyUnitPath(unitSample, gameModel.getActualPlayer().getCastle().getPosition(),
                (int)checkPoint.getX(), (int)checkPoint.getY());

        Assertions.assertFalse(boughtUnit.getCheckPoints().contains(checkPoint));
    }

    @Test
    void refreshUnitsPath() {
        Unit unitSample = new Runner(new Point2D(0,0), "test.png");
        Unit boughtUnit = gameModel.playerBuyUnit(unitSample, gameModel.getActualPlayer().getCastle().getPosition());

        Point2D checkPoint = getRandomOpenFieldFromOwnHalf(PlayerType.PLAYER_A);

        gameModel.modifyUnitPath(unitSample, gameModel.getActualPlayer().getCastle().getPosition(),
                (int)checkPoint.getX(), (int)checkPoint.getY());

        Tower tower = new Tower(new Point2D((int)checkPoint.getX(), (int)checkPoint.getY()), TowerType.BLOCKADE);
        gameModel.getMap().addTower((int)checkPoint.getX(), (int)checkPoint.getY(), tower);

        gameModel.refreshUnitsPath((int)checkPoint.getX(), (int)checkPoint.getY());

        Assertions.assertFalse(boughtUnit.getCheckPoints().contains(checkPoint));
    }

    @Test
    void initializeUnitPrioritiesTest() {
        Point2D castlePosition = gameModel.getPlayerA().getCastle().getPosition();
        Point2D barrackPosition = this.getRandomOpenFieldFromOwnHalf(PlayerType.PLAYER_A);
        Barrack barrack = new Barrack(barrackPosition);
        gameModel.getPlayerA().buyBarrack(barrack);

        gameModel.getPlayerA().buyUnit(new Demolisher(castlePosition, "test.png"));
        gameModel.getPlayerA().buyUnit(new Demolisher(castlePosition, "test.png"));
        gameModel.getPlayerA().buyUnit(new Runner(castlePosition, "test.png"));
        gameModel.getPlayerA().buyUnit(new Runner(castlePosition, "test.png"));

        gameModel.getPlayerA().buyUnit(new Demolisher(barrackPosition, "test.png"));
        gameModel.getPlayerA().buyUnit(new Demolisher(barrackPosition, "test.png"));
        gameModel.getPlayerA().buyUnit(new Runner(barrackPosition, "test.png"));
        gameModel.getPlayerA().buyUnit(new Runner(barrackPosition, "test.png"));

        gameModel.initializeUnitPriorities();

        List<Unit> playerAUnits = gameModel.getPlayerA().getUnits();

        for(Unit unit : playerAUnits) {
            boolean uniquePriority = playerAUnits.stream()
                    .filter(u -> u.getPosition() == unit.getPosition() && !u.equals(unit))
                    .allMatch(u -> u.getPriority() != unit.getPriority());

            Assertions.assertTrue(uniquePriority);
        }
    }
}
