package bitsplease.mummymayhem.model;

import bitsplease.mummymayhem.model.buildings.Barrack;
import bitsplease.mummymayhem.model.buildings.Castle;
import bitsplease.mummymayhem.model.buildings.Tower;
import bitsplease.mummymayhem.model.buildings.TowerType;
import bitsplease.mummymayhem.model.units.Runner;
import bitsplease.mummymayhem.model.units.Tank;
import bitsplease.mummymayhem.model.units.Unit;
import javafx.geometry.Point2D;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

public class PlayerTest {

    private static final int INITIAL_GOLD_AMOUNT = 1000;
    private static final int DEMOLISH_PRICE = 30;

    Player player;

    @BeforeEach
    void setUp() {
        player = new Player(PlayerType.PLAYER_A, new Point2D(5, 5));
    }

    @Test
    void setGoldTest() {
        player.setGold(900);

        Assertions.assertEquals(900, player.getGold());
    }

    @Test
    void getCastleTest() {
        Castle castle = player.getCastle();

        Assertions.assertAll(
                () -> Assertions.assertNotNull(castle),
                () -> Assertions.assertEquals(new Point2D(5, 5), castle.getPosition())
        );
    }

    @Test
    void buyUnitTest() {
        Unit unit = new Tank(new Point2D(0,0), "test.png");
        player.buyUnit(unit);

        Assertions.assertAll(
                () -> Assertions.assertEquals(INITIAL_GOLD_AMOUNT-unit.getPrice(), player.getGold()),
                () -> Assertions.assertTrue(player.getUnits().contains(unit))
        );
    }

    @Test
    void buyUnitTest2() {
        player.setGold(0);
        Unit unit = new Tank(new Point2D(0,0), "test.png");
        player.buyUnit(unit);

        Assertions.assertAll(
                () -> Assertions.assertEquals(0, player.getGold()),
                () -> Assertions.assertFalse(player.getUnits().contains(unit))
        );
    }

    @Test
    void sellUnitTest() {
        Unit unit = new Tank(new Point2D(0,0), "test.png");
        Unit unit2 = new Tank(new Point2D(1,1), "test.png");

        Assertions.assertAll(
                () -> Assertions.assertTrue(player.buyUnit(unit)),
                () -> Assertions.assertTrue(player.buyUnit(unit2)),
                () -> Assertions.assertEquals(INITIAL_GOLD_AMOUNT - (unit.getPrice() + unit2.getPrice()), player.getGold()),
                () -> Assertions.assertTrue(player.getUnits().contains(unit)),
                () -> Assertions.assertTrue(player.getUnits().contains(unit2))
        );

        int currentGoldAmount = player.getGold();

        Assertions.assertAll(
                () -> Assertions.assertTrue(player.sellUnit(unit, new Point2D(0, 0))),
                () -> Assertions.assertEquals(currentGoldAmount + unit.getPrice(), player.getGold()),
                () -> Assertions.assertFalse(player.getUnits().contains(unit)),
                () -> Assertions.assertTrue(player.getUnits().contains(unit2))
        );
    }

    @Test
    void sellUnitTest2() {
        Unit unit = new Runner(new Point2D(0,0), "test.png");

        Assertions.assertAll(
                () -> Assertions.assertFalse(player.sellUnit(unit, new Point2D(0, 0))),
                () -> Assertions.assertEquals(INITIAL_GOLD_AMOUNT, player.getGold()),
                () -> Assertions.assertFalse(player.getUnits().contains(unit))
        );

    }

    @Test
    void buyTowerTest() {
        Tower tower = new Tower(TowerType.BLOCKADE);

        Assertions.assertAll(
                () -> Assertions.assertTrue(player.buyTower(tower)),
                () -> Assertions.assertEquals(INITIAL_GOLD_AMOUNT-tower.getPrice(), player.getGold()),
                () -> Assertions.assertTrue(player.searchTower(tower))
        );
    }

    @Test
    void buyTowerTest2() {
        player.setGold(0);
        Tower tower = new Tower(TowerType.BLOCKADE);

        Assertions.assertAll(
                () -> Assertions.assertFalse(player.buyTower(tower)),
                () -> Assertions.assertEquals(0, player.getGold()),
                () -> Assertions.assertFalse(player.searchTower(tower))
        );
    }

    @Test
    void upgradeTowerTest() {
        Tower tower = new Tower(TowerType.BLOCKADE);
        player.buyTower(tower);

        Assertions.assertAll(
                () -> Assertions.assertTrue(player.upgradeTower(tower)),
                () -> Assertions.assertEquals(INITIAL_GOLD_AMOUNT - (tower.getPrice()+tower.getPrice()*0.5), player.getGold()),
                () -> Assertions.assertEquals(2, tower.getLevel())
        );
    }

    @Test
    void upgradeTowerTest2() {
        Tower tower = new Tower(TowerType.BLOCKADE);
        player.buyTower(tower);
        player.setGold(0);

        Assertions.assertAll(
                () -> Assertions.assertFalse(player.upgradeTower(tower)),
                () -> Assertions.assertEquals(0, player.getGold()),
                () -> Assertions.assertEquals(1, tower.getLevel())
        );
    }

    @Test
    void removeTowerTest() {
        Tower tower = new Tower(TowerType.BLOCKADE);
        player.buyTower(tower);
        int currentGoldAmount = player.getGold();
        player.removeTower(tower);

        Assertions.assertAll(
                () -> Assertions.assertEquals(currentGoldAmount + tower.getPrice()*0.5, player.getGold()),
                () -> Assertions.assertFalse(player.searchTower(tower))
        );
    }

    @Test
    void removeTowerTest2() {
        Tower tower = new Tower(TowerType.BLOCKADE);
        player.buyTower(tower);
        tower.takeDamage(tower.getHp());
        int currentGoldAmount = player.getGold();
        player.removeTower(tower);

        Assertions.assertAll(
                () -> Assertions.assertEquals(currentGoldAmount - DEMOLISH_PRICE, player.getGold()),
                () -> Assertions.assertFalse(player.searchTower(tower))
        );
    }

    @Test
    void searchTowerTest() {
        Tower tower = new Tower(TowerType.BLOCKADE);
        Tower tower2 = new Tower(TowerType.BLOCKADE);
        player.buyTower(tower);

        Assertions.assertAll(
                () -> Assertions.assertTrue(player.searchTower(tower)),
                () -> Assertions.assertFalse(player.searchTower(tower2))
        );
    }

    @Test
    void buyBarrackTest() {
        Barrack barrack = new Barrack(new Point2D(2, 2));

        Assertions.assertAll(
                () -> Assertions.assertTrue(player.buyBarrack(barrack)),
                () -> Assertions.assertEquals(INITIAL_GOLD_AMOUNT-Barrack.getPrice(), player.getGold()),
                () -> Assertions.assertTrue(player.getBarracks().contains(barrack))
        );
    }

    @Test
    void buyBarrackTest2() {
        Barrack barrack = new Barrack(new Point2D(2, 2));
        player.setGold(0);

        Assertions.assertAll(
                () -> Assertions.assertFalse(player.buyBarrack(barrack)),
                () -> Assertions.assertEquals(0, player.getGold()),
                () -> Assertions.assertFalse(player.getBarracks().contains(barrack))
        );
    }

    @Test
    void hasBarrackOnPositionTest() {
        Barrack barrack = new Barrack(new Point2D(1, 2));
        player.buyBarrack(barrack);

        Assertions.assertAll(
                () -> Assertions.assertTrue(player.hasBarrackOnPosition(1, 2)),
                () -> Assertions.assertFalse(player.hasBarrackOnPosition(2, 1))
        );
    }
}
