package bitsplease.mummymayhem.model.map;

import bitsplease.mummymayhem.model.buildings.*;
import javafx.geometry.Point2D;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.util.List;

class MapTest {
    Map map;

    @BeforeEach
    void setUp() {
        map = new Map(10, 10);
    }

    @Test
    void mapGenerationTest() {
        int open = 0;
        for(List<Field> row: map.getFields()) {
            for(Field f : row) {
                if(f.getType().equals(FieldType.OPEN)) {
                    open++;
                }
            }
        }

        Assertions.assertEquals(80, open);
        Assertions.assertTrue(map.getPathFinder().isTraversable());
    }

    @Test
    void setWidth() {
        Map newMap = new Map();
        newMap.setWidth(Map.MIN_COLUMN_NUMBER);
        Assertions.assertEquals(Map.MIN_COLUMN_NUMBER, newMap.getWidth());

        newMap.setWidth(Map.MIN_COLUMN_NUMBER-1);
        Assertions.assertEquals(0, newMap.getWidth());
    }

    @Test
    void setHeight() {
        Map newMap = new Map();
        newMap.setHeight(Map.MIN_ROW_NUMBER);
        Assertions.assertEquals(Map.MIN_ROW_NUMBER, newMap.getHeight());

        newMap.setHeight(Map.MIN_ROW_NUMBER-1);
        Assertions.assertEquals(0, newMap.getHeight());
    }

    @Test
    void nearCastle() {
        int x = (int)map.getPlayerACastlePosition().getX();
        int y = (int)map.getPlayerACastlePosition().getY();

        Assertions.assertTrue(map.nearCastle(x,y));
        x++;
        y--;
        Assertions.assertTrue(map.nearCastle(x,y));
        y += 5;
        Assertions.assertFalse(map.nearCastle(x,y));

        x = (int)map.getPlayerBCastlePosition().getX();
        y = (int)map.getPlayerBCastlePosition().getY();

        Assertions.assertTrue(map.nearCastle(x,y));
        x--;
        y++;
        Assertions.assertTrue(map.nearCastle(x,y));
        x -= 4;
        Assertions.assertFalse(map.nearCastle(x,y));
    }

    @Test
    void isCastlePosition() {
        int x = (int)map.getPlayerACastlePosition().getX();
        int y = (int)map.getPlayerACastlePosition().getY();
        Assertions.assertTrue(map.isCastlePosition(x,y));

        x = (int)map.getPlayerBCastlePosition().getX();
        y = (int)map.getPlayerBCastlePosition().getY();
        Assertions.assertTrue(map.isCastlePosition(x,y));

        x+=5;
        y+=2;
        Assertions.assertFalse(map.isCastlePosition(x,y));
    }

    @Test
    void getPlayerAFieldsPositions() {
        Point2D pos = new Point2D(3,8);
        Assertions.assertTrue(map.getPlayerAFieldsPositions().contains(pos));

        Point2D pos2 = new Point2D(6,4);
        Assertions.assertFalse(map.getPlayerAFieldsPositions().contains(pos2));
    }

    @Test
    void getPlayerBFieldsPositions() {
        Point2D pos = new Point2D(3,8);
        Assertions.assertFalse(map.getPlayerBFieldsPositions().contains(pos));

        Point2D pos2 = new Point2D(6,4);
        Assertions.assertTrue(map.getPlayerBFieldsPositions().contains(pos2));
    }

    @Test
    void getPlayerAPossibleBarrackPositions() {
        Point2D pos = new Point2D(2,8);
        Assertions.assertTrue(map.getPlayerAPossibleBarrackPositions().contains(pos));

        Point2D pos2 = new Point2D(3,4);
        Assertions.assertFalse(map.getPlayerAPossibleBarrackPositions().contains(pos2));
    }

    @Test
    void getPlayerBPossibleBarrackPositions() {
        Point2D pos = new Point2D(8,8);
        Assertions.assertTrue(map.getPlayerBPossibleBarrackPositions().contains(pos));

        Point2D pos2 = new Point2D(5,4);
        Assertions.assertFalse(map.getPlayerBPossibleBarrackPositions().contains(pos2));
    }

    @Test
    void blockPathTest() {
        for(int i = 0; i< 10; i++) {
            for(int j = 0; j < 10; j++) {
                if(i == 4 && j != 5) {
                    map.setField(i,j,FieldType.STONE);
                }
                else {
                    map.setField(i,j,FieldType.OPEN);
                }
            }
        }

        map.changePathFinder();
        Building barrack = new Barrack(new Point2D(4,5));
        Assertions.assertTrue(map.blockPath(barrack));

        barrack.setPosition(map.getPlayerACastlePosition());
        Assertions.assertTrue(map.blockPath(barrack));

        barrack.setPosition(new Point2D(3,1));
        Assertions.assertFalse(map.blockPath(barrack));
    }

    @Test
    void addTowerTest() {
        Tower tower = new Tower(new Point2D(3,3), TowerType.BLOCKADE);
        map.getField(3,3).addTower(tower);
        Assertions.assertEquals(tower,map.getField(3,3).getTower());
    }

    @Test
    void addBuildingTest() {
        Castle c = new Castle(new Point2D(4,4),"");
        map.getField(4,4).addBuilding(c);
        Assertions.assertEquals(c,map.getField(4,4).getBuilding());
    }

    @Test
    void testToString() {
        String expected = """
                OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN\s
                OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN\s
                OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN\s
                OPEN OPEN OPEN OPEN OPEN DUNE OPEN OPEN OPEN OPEN\s
                OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN\s
                OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN\s
                OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN\s
                OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN\s
                OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN\s
                OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN OPEN\s
                """;

        for(int i = 0; i< 10; i++) {
            for(int j = 0; j < 10; j++) {
                map.setField(i,j,FieldType.OPEN);
            }
        }
        map.setField(3,5, FieldType.DUNE);

        Assertions.assertEquals(expected, map.toString());
    }
}