package bitsplease.mummymayhem.model.units;

import bitsplease.mummymayhem.model.buildings.Castle;
import bitsplease.mummymayhem.model.buildings.Tower;
import bitsplease.mummymayhem.model.buildings.TowerType;
import bitsplease.mummymayhem.model.map.Field;
import bitsplease.mummymayhem.model.map.FieldType;
import bitsplease.mummymayhem.model.map.Map;
import bitsplease.mummymayhem.model.utils.Direction;
import javafx.geometry.Point2D;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

public class UnitTest {
    Unit unit;

    @BeforeEach
    void setUp() {
        unit = new Fighter(new Point2D(0,0), "src/main/resources/assets/units/fighter/");
        unit.initAnimationStages();
        unit.setFacing(Direction.RIGHT);
    }

    @Test
    void takeDamageTest() {
        unit.setHp(100);
        unit.takeDamage(15);
        Assertions.assertEquals(85, unit.getHp());
    }

    @Test
    void takeDamageTest2() {
        unit.setHp(5);
        unit.takeDamage(15);
        Assertions.assertEquals(0, unit.getHp());
    }

    @Test
    void attackTest() {
        Tower tower = new Tower(TowerType.BLOCKADE);

        tower.setHp(100);
        unit.setAttack(15);
        unit.attack(tower);

        Assertions.assertEquals(85, tower.getHp());
    }

    @Test
    void attackTest2() {
        Tower tower = new Tower(TowerType.BLOCKADE);

        tower.setHp(5);
        unit.setAttack(15);
        unit.attack(tower);

        Assertions.assertEquals(0, tower.getHp());
    }

    @Test
    void attackTest3() {
        Castle castle = new Castle(new Point2D(0,0), "");

        castle.setHp(100);
        unit.setAttack(15);
        unit.attack(castle);

        Assertions.assertEquals(85, castle.getHp());
    }

    @Test
    void attackTest4() {
        Castle castle = new Castle(new Point2D(0,0), "");

        castle.setHp(5);
        unit.setAttack(15);
        unit.attack(castle);

        Assertions.assertEquals(0, castle.getHp());
    }

    @Test
    void moveXTest() {
        unit.moveX(1);

        Assertions.assertEquals(new Point2D(1,0),unit.getPosition());
    }

    @Test
    void moveYTest() {
        unit.moveY(1);

        Assertions.assertEquals(new Point2D(0,1),unit.getPosition());
    }

    @Test
    void calculateDefaultPathTest() {
        Map map = new Map(10,10);
        unit.setPosition(map.getPlayerACastlePosition());

        unit.calculateDefaultPath(map, map.getPlayerBCastlePosition());

        Assertions.assertTrue(unit.getPath().size() > 0);

        for(Field f: unit.getPath()) {
            Assertions.assertEquals(FieldType.OPEN, f.getType());
        }
    }

    @Test
    void getUpdatedPathTest() {
        Map map = new Map(10,10);

        for (List<Field> fields : map.getFields()) {
            for (Field field : fields) {
                field.setType(FieldType.OPEN);
            }
        }

        unit.setPosition(map.getPlayerACastlePosition());

        unit.calculateDefaultPath(map, map.getPlayerBCastlePosition());
        Random rand = new Random();

        // Generates 3 random checkpoints
        for(int i = 0; i < 3; i++) {
            int randX;
            int randY;
            do {
                randX = rand.nextInt(10);
                randY = rand.nextInt(10);

            } while(map.getField(randY, randX).getType()!= FieldType.OPEN
                    || map.isCastlePosition(randX,randY)
                    || unit.getCheckPoints().contains(new Point2D(randX,randY)));

            unit.addCheckPoint(new Point2D(randX, randY));
        }

        Assertions.assertTrue(unit.getPath().size() > 0);

        int cpInUpdatedPath = 0;

        for(Point2D p: unit.getUpdatedPath(map)) {
            Assertions.assertEquals(FieldType.OPEN, map.getField((int)p.getY(), (int)p.getX()).getType());
            Assertions.assertFalse(map.getField((int)p.getY(), (int)p.getX()).hasTower());

            for (Point2D cp: unit.getCheckPoints()) {
                if(p.equals(cp)) {
                    cpInUpdatedPath++;
                }
            }
        }

        Assertions.assertTrue(cpInUpdatedPath >= 3);
    }

    @Test
    void nextFieldDirectionTest() {
        unit.setPosition(new Point2D(1,1));

        List<Field> fields = new ArrayList<>();
        fields.add(new Field(new Point2D(1,0), FieldType.OPEN));
        fields.add(new Field(new Point2D(2,1), FieldType.OPEN));
        fields.add(new Field(new Point2D(0,1), FieldType.OPEN));
        fields.add(new Field(new Point2D(1,2), FieldType.OPEN));

        unit.setPath(fields);

        Assertions.assertEquals(Direction.UP, unit.nextFieldDirection());
        unit.getPath().remove(0);

        Assertions.assertEquals(Direction.RIGHT, unit.nextFieldDirection());
        unit.getPath().remove(0);

        Assertions.assertEquals(Direction.LEFT, unit.nextFieldDirection());
        unit.getPath().remove(0);

        Assertions.assertEquals(Direction.DOWN, unit.nextFieldDirection());
        unit.getPath().remove(0);
    }

    @Test
    void nextAnimationPhaseTest() {
        Assertions.assertEquals(0, unit.getCurrentAnimationPhase());
        Assertions.assertEquals(1, unit.getNextAnimationPhase());

        unit.nextAnimationPhase();

        Assertions.assertEquals(1, unit.getCurrentAnimationPhase());
        Assertions.assertEquals(2, unit.getNextAnimationPhase());

        unit.nextAnimationPhase();

        Assertions.assertEquals(2, unit.getCurrentAnimationPhase());
        Assertions.assertEquals(1, unit.getNextAnimationPhase());

        unit.nextAnimationPhase();

        Assertions.assertEquals(1, unit.getCurrentAnimationPhase());
        Assertions.assertEquals(0, unit.getNextAnimationPhase());

        unit.nextAnimationPhase();

        Assertions.assertEquals(0, unit.getCurrentAnimationPhase());
        Assertions.assertEquals(1, unit.getNextAnimationPhase());
    }

    @Test
    void checkNextFieldTest() {
        unit.setPosition(new Point2D(1,1));
        unit.addFieldToPath(new Field(new Point2D(2,1), FieldType.OPEN));
        unit.addFieldToPath(new Field(new Point2D(2,2), FieldType.OPEN));
        unit.addFieldToPath(new Field(new Point2D(1,2), FieldType.OPEN));
        unit.addFieldToPath(new Field(new Point2D(1,1), FieldType.OPEN));

        unit.checkNextField();
        Assertions.assertEquals(Direction.RIGHT, unit.getFacing());

        unit.setPosition(unit.getPath().get(0).getPosition());
        unit.checkNextField();
        Assertions.assertEquals(Direction.DOWN, unit.getFacing());

        unit.setPosition(unit.getPath().get(0).getPosition());
        unit.checkNextField();
        Assertions.assertEquals(Direction.LEFT, unit.getFacing());

        unit.setPosition(unit.getPath().get(0).getPosition());
        unit.checkNextField();
        Assertions.assertEquals(Direction.UP, unit.getFacing());
    }

    @Test
    void isDestinationReachedTest() {
        Map map = new Map(10,10);
        unit.setPosition(map.getPlayerACastlePosition());
        unit.calculateDefaultPath(map, map.getPlayerBCastlePosition());

        Assertions.assertFalse(unit.isDestinationReached());

        //Checking when unit is right on the destination
        unit.setPosition(map.getPlayerBCastlePosition());
        Assertions.assertTrue(unit.isDestinationReached());

        //Checking when unit is going right an overshoots it a bit
        unit.moveX(0.1);
        Assertions.assertTrue(unit.isDestinationReached());

        //Checking when unit is going left and is on the right side of the destination
        unit.setFacing(Direction.LEFT);
        Assertions.assertFalse(unit.isDestinationReached());

        //Checking when unit is going left and overshoots it a bit
        unit.moveX(-0.2);
        Assertions.assertTrue(unit.isDestinationReached());
        unit.moveX(0.2);

        unit.setFacing(Direction.DOWN);

        //Checking when unit is going down and overshoots it a bit
        unit.moveY(0.2);
        Assertions.assertTrue(unit.isDestinationReached());


        //Checking when unit is going up and is to the south of the destination
        unit.setFacing(Direction.UP);
        Assertions.assertFalse(unit.isDestinationReached());

        //Checking when unit is going up and overshoots it a bit
        unit.moveY(-0.2);
        Assertions.assertTrue(unit.isDestinationReached());
    }
}
